# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_129_273
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 129 273.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..etsi_3gpp.ts_129_061 import X3gppChargingCharacteristicsAVP
from ..etsi_3gpp.ts_129_212 import RatTypeAVP
from ..etsi_3gpp.ts_129_272 import AmbrAVP
from ..etsi_3gpp.ts_129_272 import ContextIdentifierAVP
from ..etsi_3gpp.ts_129_272 import ApnConfigurationAVP

from ..ietf.rfc4006 import SubscriptionIdAVP
from ..ietf.rfc6733 import SessionTimeoutAVP

from ...base import DiameterAVP
from ...constants.etsi_3gpp.ts_129_273 import *
from ...types import *


class MobileNodeIdentifierAVP(DiameterAVP, UTF8StringType):
    """Implementation of Mobile-Node-Identifier AVP in Section 5.2.3.2  
    of ETSI TS 129 273 V15.4.0 (2019-10).

    The Mobile-Node-Identifier AVP (AVP Code 506) is of type UTF8String.
    """
    code = MOBILE_NODE_IDENTIFIER_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             MobileNodeIdentifierAVP.code,
                             MobileNodeIdentifierAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)
        

class Non3gppIpAccessAVP(DiameterAVP, EnumeratedType):
    """Implementation of Non-3GPP-IP-Access AVP in Section 8.2.3.3 
    of ETSI TS 129 273 V14.5.0 (2019-10).

    The Non-3GPP-IP-Access AVP (AVP Code 1501) is of type Enumerated.
    """
    code = NON_3GPP_IP_ACCESS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                NON_3GPP_SUBSCRIPTION_ALLOWED,
                NON_3GPP_SUBSCRIPTION_BARRED
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             Non3gppIpAccessAVP.code,
                             Non3gppIpAccessAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class Mip6FeatureVectorAVP(DiameterAVP, Unsigned64Type):
    """Implementation of MIP6-Feature-Vector AVP in Section 9.2.3.2.3 of 
    ETSI TS 129 273 V14.3.0 (2017-07).

    The MIP6-Feature-Vector AVP (AVP Code 124) is of type Unsigned64.
    """
    code = MIP6_FEATURE_VECTOR_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, Mip6FeatureVectorAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned64Type.__init__(self, data=data)


class Non3gppIpAccessApnAVP(DiameterAVP, EnumeratedType):
    """Implementation of Non-3GPP-Ip-Access-APN AVP in Section 8.2.3.4
    of ETSI TS 129 273 V14.3.0 (2017-07).

    The Non-3GPP-Ip-Access-APN AVP (AVP Code 1502) is of type Enumerated.
    """
    code = NON_3GPP_IP_ACCESS_APN_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                NON_3GPP_APNS_ENABLE,
                NON_3GPP_APNS_DISABLE,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             Non3gppIpAccessApnAVP.code,
                             Non3gppIpAccessApnAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class Non3gppUserDataAVP(DiameterAVP, GroupedType):
    """Implementation of Non-3GPP-User-Data AVP in Section 8.2.3.1 
    of ETSI TS 129 273 V14.3.0 (2017-07).

    The Non-3GPP-User-Data AVP (AVP Code 1500) is of type Grouped.
    """
    code = NON_3GPP_USER_DATA_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    optionals = {
                    "subscription_id": SubscriptionIdAVP,
                    "non_3gpp_ip_access": Non3gppIpAccessAVP,
                    "non_3gpp_ip_access_apn": Non3gppIpAccessApnAVP,
                    "rat_type": RatTypeAVP,
                    "session_timeout": SessionTimeoutAVP,
                    "mip6_feature_vector": Mip6FeatureVectorAVP,
                    "ambr": AmbrAVP,
                    "x3gpp_charging_characteristics": X3gppChargingCharacteristicsAVP,
                    "context_identifier": ContextIdentifierAVP,
                    #"apn_oi_replacement": ApnOiReplacementAVP,
                    "apn_configuration": ApnConfigurationAVP,
                    #"trace_info": TraceInfoAVP,
                    #"twan_default_apn_context_id": TwanDefaultApnContextIdAVP,
                    #"twan_access_info": TwanAccessInfoAVP,
                    #"emergency_info": EmergencyInfoAVP,
                    #"erp_authorization": ErpAuthorizationAVP

    }

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             Non3gppUserDataAVP.code,
                             Non3gppUserDataAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AnTrustedAVP(DiameterAVP, EnumeratedType):
    """Implementation of AN-Trusted AVP in Section 5.2.3.9 of
    ETSI TS 129 273 V12.5.0 (2014-10).

    The AN-Trusted AVP (AVP Code 1503) is of type Enumerated.
    """
    code = AN_TRUSTED_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                AN_TRUSTED_TRUSTED,
                AN_TRUSTED_UNTRUSTED,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AnTrustedAVP.code,
                             AnTrustedAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)



# -*- coding: utf-8 -*-
"""
    bromelia.base
    ~~~~~~~~~~~~~

    This module contains the base structures that power 
    bromelia.messages module.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import os
import re
from copy import deepcopy

from ._internal_utils import avp_look_up
from ._internal_utils import header_representation
from .constants import *
from .exceptions import AVPAttributeValueError
from .exceptions import AVPParsingError
from .exceptions import DiameterAvpError
from .exceptions import DiameterHeaderAttributeValueError
from .exceptions import DiameterHeaderError
from .exceptions import DiameterMessageError
from .utils import is_vendor_id


class DiameterAvpLoader:
    """Helper class used to load all available DiameterAVP subclasses
    defined in the Bromelia library. It supports the DiameterAVP's 
    constructor and its load staticmethod. The former helps specifying 
    the DiameterAVP __repr__ dunder method and the latter supports the
    instantiation of specialized DiameterAVP objects on the go.

    Specialized DiameterAVP objects refer to DiameterAVP subclasses 
    objects.

    DiameterAvpLoader class is expected to be used only inside the 
    Bromelia library implementation. There is no public API to be
    exposed to third-party.
    """

    def __init__(self):
        self.avps = None


    def has_updated(self):
        avps = DiameterAVP.__subclasses__()
        if len(self.avps) == len(avps):
            return False
        return True


    def _get_load_avps_dictionary(self):
        avps = DiameterAVP.__subclasses__()
        loaded_avps = dict()
        for avp in avps:
            if avp.vendor_id is not None:
                if avp.vendor_id in loaded_avps:
                    loaded_avps[avp.vendor_id].update({avp.code: avp})
                else:
                    loaded_avps.update({avp.vendor_id: {avp.code: avp}})
            else:
                if VENDOR_ID_DEFAULT in loaded_avps:
                    loaded_avps[VENDOR_ID_DEFAULT].update({avp.code: avp})
                else:
                    loaded_avps.update({VENDOR_ID_DEFAULT: {avp.code: avp}})

        return loaded_avps


    def get_avp_class(self, avp):
        if self.avps is not None:
            if self.has_updated():
                self.avps = self._get_load_avps_dictionary()
        else:
            self.avps = self._get_load_avps_dictionary()

        if avp.vendor_id is not None: 
            return self.avps[avp.vendor_id][avp.code]
        
        return self.avps[VENDOR_ID_DEFAULT][avp.code]


    def get_avp_class_name(self, avp):
        try: 
            avp_name = self.get_avp_class(avp).__name__[:-3]
            if len(avp_name.split("-")) == 1:
                return "-".join(re.findall("[A-Z][^A-Z]*", avp_name))

        except KeyError:
            return "Unknown"    


    def _get_avp_class_name(self, avp):
        try: 
            avp_name = self.get_avp_class(avp).__name__[:-3]
            if len(avp_name.split("-")) == 1:
                return "_".join(re.findall("[A-Z][^A-Z]*", avp_name)).lower() + "_avp"

        except KeyError:
            return "Unknown"    


class DiameterAVP(object):
    """Implementation of a Diameter AVP. 
    
    Refer to Section 4 of IETF RFC 6733 for details. The constructor has
    input arguments as per Header AVP fields.

    Provides a general-case interface for create custom and standard 
    Diameter AVPs.

    :param code: the AVP Code which identifies the attribute uniquely.
    :param flags: the AVP Flags field informs the receiver how each attribute
        must be handled.
    :param length: the AVP Length field is three octets, and indicates the 
        number of octets in this AVP including the AVP Code field, AVP Length
        field, AVP Flags field, Vendor-ID field (if present), and the AVP Data
        field.
    :param vendor_id: the Vendor-ID field is present if the 'V' bit is set
        in the AVP Flags field. It is the only optional field in a AVP header.
    :param data: the Data field is zero or more octets and contains 
        information specific to the Attribute.

    Usage::

        >>> from bromelia.avps import DiameterAVP
        >>> avp = DiameterAVP()
        >>> avp
        <Diameter AVP: 0 [Diameter]>
    """

    __slots__ = ("_code", "_flags", "_length", 
                 "_vendor_id", "_data", "_padding")

    flag_vendor_id_bit = convert_to_1_byte(0x80)
    flag_mandatory_bit = convert_to_1_byte(0x40)
    flag_protected_bit = convert_to_1_byte(0x20)
    flag_reserved5_bit = convert_to_1_byte(0x10)
    flag_reserved4_bit = convert_to_1_byte(0x08)
    flag_reserved3_bit = convert_to_1_byte(0x04)
    flag_reserved2_bit = convert_to_1_byte(0x02)
    flag_reserved1_bit = convert_to_1_byte(0x01)

    def __init__(self, 
                 code=0, 
                 vendor_id=None, 
                 flags=0, 
                 data=None, 
                 padding=None):
        self.code = code
        self.vendor_id = vendor_id
        self.flags = flags
        self.data = data
        self._padding = padding


    def __repr__(self):
        """Returns a DiameterAVP object representation in a format which 
        identify the Diameter AVP code and the Diameter AVP name.
        """
        avp_name = loader.get_avp_class_name(self)
        if avp_name == "Unknown":
            avp_name = avp_look_up(self)

        flag_representation = ""

        if self.is_vendor_id():
            flag_representation += " VENDOR,"

        if self.is_mandatory():
            flag_representation += " MANDATORY,"

        if self.is_protected():
            flag_representation += " PROTECTED "

        return f"<Diameter AVP: {self.get_code()} [{avp_name}]"\
                f"{flag_representation[:-1]}>"


    def __add__(self, other):
        """Dunder method to concatenate two DiameterAVP objects and return 
        a byte stream representing those two DiameterAVP objects.
        """
        return self.dump() + other.dump()


    def __len__(self):
        """Dunder method to get the length of a DiameterAVP object as per 
        the AVP Length field.
        """
        return self.get_length()


    def __eq__(self, other):
        """Dunder method to check if the byte stream of two DiameterAVP objects 
        are equal.
        """
        return self.dump() == other.dump()


    def __bytes__(self):
        """Dunder method to return the byte stream of a DiameterAVP object. It
        is the dump() method used in another way.
        """
        return self.dump()


    def copy(self):
        """Creates a deepcopy of a DiameterAVP object.
        """
        return deepcopy(self)


    @classmethod
    def convert(cls, avp):
        """Used to convert a specialized DiameterAVP object into a generic 
        DiameterAVP object.
        """
        if not isinstance(avp, DiameterAVP):
            raise DiameterAvpError("invalid AVP. It MUST be a DiameterAVP "\
                                   "subclass object to be converted into "\
                                   "DiameterAVP object")
        
        return cls(code=avp.code,
                   vendor_id=avp.vendor_id,
                   flags=avp.flags,
                   data=avp.data,
                   padding=avp._padding)


    @property
    def code(self):
        """Getter to AVP Code field.
        """
        return self._code


    @code.setter
    def code(self, value):
        """Setter to AVP Code field.
        """
        if isinstance(value, int) and (value >= 0 and value <= 4294967295):
            self._code = convert_to_4_bytes(value)

        if isinstance(value, bytes) and len(value) == 4:
            self._code = value

        elif isinstance(value, int) and (value < 0 or value > 4294967295):
            raise AVPAttributeValueError("code attribute has 4-bytes "\
                                         "length long")

        elif isinstance(value, bytes) and len(value) != 4:
            raise AVPAttributeValueError("code attribute has 4-bytes "\
                                         "length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise AVPAttributeValueError("invalid code attribute value")


    @property
    def flags(self):
        """Getter to AVP Flags field.
        """
        return self._flags


    @flags.setter
    def flags(self, value):
        """Setter to AVP Flags field.
        """
        if isinstance(value, int) and (value >= 0 and value <= 255):
            self._flags = convert_to_1_byte(value)

        elif isinstance(value, bytes) and len(value) == 1:
            self._flags = value

        elif isinstance(value, int) and (value < 0 or value > 255):
            raise AVPAttributeValueError("flags attribute has 1-byte "\
                                         "length long")

        elif isinstance(value, bytes) and len(value) != 1:
            raise AVPAttributeValueError("flags attribute has 1-byte "\
                                         "length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise AVPAttributeValueError("invalid flags attribute value")


    @property
    def length(self):
        """Getter to AVP Length field.
        """
        try:
            length = len(self.data)
        except TypeError as e:
            if "object of type 'NoneType' has no len()" in e.args[0]:
                length = 0

        if self.vendor_id:        
            length += AVP_HEADER_LENGTH_LONGER

        else:
            length += AVP_HEADER_LENGTH

        return convert_to_3_bytes(length)


    @length.setter
    def length(self, value):
        """Setter to AVP Length field.
        """
        if value is None:
            self._length = value

        elif isinstance(value, int) and (value >= 0 and value <= 16777215):
            self._length = convert_to_3_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 16777215):
            raise AVPAttributeValueError("length attribute has 3-bytes length "\
                                         "long")

        elif isinstance(value, bytes) and len(value) == 3:
            self._length = value

        elif isinstance(value, bytes) and len(value) != 3:
            raise AVPAttributeValueError("length attribute has 3-bytes length "\
                                         "long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise AVPAttributeValueError("invalid length attribute value")

    @property
    def vendor_id(self):
        """Getter to AVP Vendor Id field.
        """
        return self._vendor_id


    @vendor_id.setter
    def vendor_id(self, value):
        """Setter to AVP Vendor Id field.
        """
        if value is None:
            self._vendor_id = value

        elif isinstance(value, int) and (value >= 0 and value <= 4294967295):
            self._vendor_id = convert_to_4_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 4294967295):
            raise AVPAttributeValueError("vendor_id attribute has 4-bytes "\
                                         "length long")

        elif isinstance(value, bytes) and len(value) == 4:
            self._vendor_id = value

        elif isinstance(value, bytes) and len(value) != 4:
            raise AVPAttributeValueError("vendor_id attribute has 4-bytes "\
                                        "length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise AVPAttributeValueError("invalid vendor_id attribute value")


    @property
    def data(self):
        """Getter to AVP Data content.
        """
        return self._data


    @data.setter
    def data(self, value):
        """Setter to AVP Data content.
        """
        data = value
        if value:
            if isinstance(value, str):
                data = value.encode("utf-8")
            elif isinstance(value, int):
                data = convert_to_4_bytes(value)
            elif isinstance(value, bytes):
                data = value
            else:
                raise AVPAttributeValueError("invalid data value. Consider "\
                                             "create a data type object "\
                                             "before assigning")

        self._data = data                                


    @property
    def padding(self):
        """Getter to AVP padding. It calculates the AVP padding if needed. 
        Otherwise it returns None.
        """
        if self.data:
            mod = len(self.data) % 4
            if mod != 0:
                return bytes(4 - mod)

        return None


    @staticmethod
    def load(stream):
        """Load a byte stream which represents Diameter AVPs and returns a 
        list of specialized DiameterAVP objects. In case it is not possible
        to identify the Diameter AVP represented by the byte stream, it should
        return a list of DiameterAVP objects.
        """
        avps = list()
        index = 0
        
        while index < len(stream):
            avp = DiameterAVP()

            try:
                avp._code = stream[index:index+4]
                avp._flags = convert_to_1_byte(stream[index+4])
                avp._length = stream[index+5:index+8]
                if len(avp._length) != 3:
                    raise AVPParsingError("invalid bytes stream. It "\
                                    "contains only the code and flags fields")

                boundary = int.from_bytes(avp._length, byteorder="big")
                
                if is_vendor_id(avp._flags):
                    avp._vendor_id = stream[index+8:index+12]
                    if len(avp._vendor_id) != 4:
                        raise AVPParsingError("invalid bytes stream. It "\
                                        "contains only the code, flags "\
                                        "and length fields")

                    avp._data = stream[index+12:index+boundary]
                    avp_header_length = AVP_HEADER_LENGTH_LONGER

                else:
                    avp._vendor_id = None
                    avp._data = stream[index+8:index+boundary]
                    avp_header_length = AVP_HEADER_LENGTH

                if len(avp._data) != boundary - avp_header_length:
                        raise AVPParsingError("invalid bytes stream. The "\
                                    "length field value does not correspond "\
                                    "to the AVP length")


            except IndexError as e:
                if e.args[0] == "index out of range":
                    raise AVPParsingError("invalid bytes stream")


            if (boundary % 4) != 0:
                padding = 4 - (boundary % 4)
            else:
                padding = 0

            avp._padding = stream[index+boundary:index+boundary+padding]

            index += boundary + padding

            try:
                _avp_class = loader.get_avp_class(avp)
                avp_object = _avp_class(avp.data)
                avps.append(avp_object)

            except KeyError as e:
                avps.append(avp)

        return avps


    def set_vendor_id_bit(self, state):
        """Set / unset the vendor id bit in DiameterAVP object.
        """
        if not isinstance(state, bool):
            raise AVPAttributeValueError("V-bit is of type Boolean")

        if state:
            if self.is_vendor_id():
                raise AVPAttributeValueError("V-bit was already set")

            flags = (self.get_flags() | 
                     DiameterAVP.get_flags_bit(DiameterAVP.flag_vendor_id_bit))

        else:
            if not self.is_vendor_id():
                raise AVPAttributeValueError("V-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterAVP.get_flags_bit(DiameterAVP.flag_vendor_id_bit))

        self.flags = convert_to_1_byte(flags)


    def is_vendor_id(self):
        """Function to check if DiameterAVP object has vendor id bit enabled.
        """
        return (self.get_flags() & 
                DiameterAVP.get_flags_bit(DiameterAVP.flag_vendor_id_bit) != 0)


    def set_mandatory_bit(self, state):
        """Set / unset the mandatory bit in DiameterAVP object.
        """
        if not isinstance(state, bool):
            raise AVPAttributeValueError("M-bit is of type Boolean")

        if state:
            if self.is_mandatory():
                raise AVPAttributeValueError("M-bit was already set")

            flags = (self.get_flags() | 
                     DiameterAVP.get_flags_bit(DiameterAVP.flag_mandatory_bit))
        else:
            if not self.is_mandatory():
                raise AVPAttributeValueError("M-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterAVP.get_flags_bit(DiameterAVP.flag_mandatory_bit))

        self.flags = convert_to_1_byte(flags)


    def is_mandatory(self):
        """Function to check if DiameterAVP object has mandatory bit enabled.
        """
        return (self.get_flags() & 
                DiameterAVP.get_flags_bit(DiameterAVP.flag_mandatory_bit) != 0)


    def set_protected_bit(self, state):
        """Set / unset the protected bit in DiameterAVP object.
        """
        if not isinstance(state, bool):
            raise AVPAttributeValueError("P-bit is of type Boolean")

        if state:
            if self.is_protected():
                raise AVPAttributeValueError("P-bit was already set")

            flags = (self.get_flags() | 
                     DiameterAVP.get_flags_bit(DiameterAVP.flag_protected_bit))
        else:
            if not self.is_protected():
                raise AVPAttributeValueError("P-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterAVP.get_flags_bit(DiameterAVP.flag_protected_bit))

        self.flags = convert_to_1_byte(flags)


    def is_protected(self):
        """Function to check if DiameterAVP object has protected bit enabled.
        """
        return (self.get_flags() & 
                DiameterAVP.get_flags_bit(DiameterAVP.flag_protected_bit) != 0)


    def get_code(self):
        """Returns the Diameter AVP code bit in Integer format.
        """
        return int.from_bytes(self.code, byteorder="big")


    def get_flags(self):
        """Returns the Diameter AVP flags bit in Integer format.
        """
        return int.from_bytes(self.flags, byteorder="big")


    def get_length(self):
        """Returns the Diameter AVP length bit in Integer format.
        """
        return int.from_bytes(self.length, byteorder="big")


    def get_vendor_id(self):
        """Returns the Diameter AVP vendor id bit in Integer format in case 
        there is vendor id. Otherwise it returns None.
        """
        if self.vendor_id:
            return int.from_bytes(self.vendor_id, byteorder="big")
        return None


    def get_padding_length(self):
        """Returns the Diameter AVP padding in Integer format in case there is
        padding. Otherwise it returns None.
        """
        if self.padding:
            return len(self.padding)
        return None


    @staticmethod
    def get_flags_bit(flag_bit):
        """Returns the Diameter AVP flags bit in Integer format of a given
        flag bit in hexadecimal value.
        """
        return int.from_bytes(flag_bit, byteorder="big")


    def dump(self):
        """Dump a byte stream which represents a DiameterAVP object serialized.
        """
        stream = self.code + self.flags + self.length
        if self.vendor_id:
            stream += self.vendor_id

        if self.data:
            stream += self.data

        if self.padding:
            stream += self.padding

        return stream


class DiameterHeader(object):
    """Implementation of a Diameter Header. 
    
    Refer to Section 3 of IETF RFC 6733 for details. The constructor has
    input arguments as per Diameter Header fields.

    Provides a general-case interface for create custom and standard Diameter 
    Header.

    WARNING: strongly not recommended passing the `length` input argument. 
    It is used for internal implementation purpose, in order to support the
    `.load()` classmethod. 

    :param version: represents the Version field to indicate Diameter Version.
    :param flags: represents the Command Flags field.
    :param command_code: represents the Command Code field.
    :param application_id: represents the Application-ID field.
    :param hop_by_hop: represents the Hop-by-Hop Identifier field.
    :param end_to_end: represents the End-to-End Identifier field.
    :param length: represents the Message Length field.

    Usage::

        >>> from bromelia.base import DiameterHeader
        >>> header = DiameterHeader()
        >>> header
        <Diameter Header: Unknown [], 0 [Diameter common message]>
    """

    __slots__ = ("_version", "_length", "_flags", "_command_code", 
                 "_application_id", "_hop_by_hop", "_end_to_end")

    flag_request_bit = convert_to_1_byte(0x80)
    flag_proxiable_bit = convert_to_1_byte(0x40)
    flag_error_bit = convert_to_1_byte(0x20)
    flag_retransmitted = convert_to_1_byte(0x10)
    flag_reserved4_bit = convert_to_1_byte(0x08)
    flag_reserved3_bit = convert_to_1_byte(0x04)
    flag_reserved2_bit = convert_to_1_byte(0x02)
    flag_reserved1_bit = convert_to_1_byte(0x01)


    def __init__(self, 
                 version=DIAMETER_VERSION,
                 flags=0,
                 command_code=0,
                 application_id=0,
                 hop_by_hop=0,
                 end_to_end=0,
                 length=20):
                
        self.version = version
        self.length = length
        self.flags = flags
        self.command_code = command_code
        self.application_id = application_id
        self.hop_by_hop = hop_by_hop
        self.end_to_end = end_to_end


    def __repr__(self):
        """Returns a DiameterHeader object representation in a format which 
        identify the Diameter code and the Diameter AVP name.
        """
        representations = header_representation(self)
    
        return "<Diameter Header: "\
                "{cmd_code_int} [{cmd_code_str}]"\
                "{flag_representation}, "\
                "{app_id_int} [{app_id_str}]>".format(**representations)


    def __len__(self):
        """Dunder method to get the length of a DiameterHeader object as per 
        the Diameter Message Length field.
        """
        return self.get_length()


    def __eq__(self, other):
        """Dunder method to check if the byte stream of two DiameterHeader 
        objects are equal.
        """
        return self.dump() == other.dump()


    def __bytes__(self):
        """Dunder method to return the byte stream of a DiameterHeader object. 
        It is the dump() method used in another way.
        """
        return self.dump()


    def copy(self):
        """Creates a deepcopy of a DiameterHeader object.
        """
        return deepcopy(self)


    @property
    def version(self):
        """Getter to Diameter Header Version field.
        """
        return self._version


    @version.setter
    def version(self, value):
        """Setter to Diameter Header Version field.
        """
        if isinstance(value, int) and (value >= 0 and value <= 255):
            self._version = convert_to_1_byte(value)

        elif isinstance(value, bytes) and len(value) == 1:
            self._version = value

        elif isinstance(value, int) and (value < 0 or value > 255):
            raise DiameterHeaderAttributeValueError("version attribute has "\
                                                    "1-byte length long")

        elif isinstance(value, bytes) and len(value) != 1:
            raise DiameterHeaderAttributeValueError("version attribute has "\
                                                    "1-byte length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid version "\
                                                    "attribute value")


    @property
    def length(self):
        """Getter to Diameter Header Length field.
        """
        return self._length


    @length.setter
    def length(self, value):
        """Setter to Diameter Header Length field.
        """
        if value is None:
            self._length = value

        elif isinstance(value, int) and (value >= 0 and value <= 16777215):
            self._length = convert_to_3_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 16777215):
            raise DiameterHeaderAttributeValueError("length attribute "\
                                                    "has 3-bytes length long")

        elif isinstance(value, bytes) and len(value) == 3:
            self._length = value

        elif isinstance(value, bytes) and len(value) != 3:
            raise DiameterHeaderAttributeValueError("length attribute "\
                                                    "has 3-bytes length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid length "\
                                                    "attribute value")


    @property
    def flags(self):
        """Getter to Diameter Header Command Flags field.
        """
        return self._flags


    @flags.setter
    def flags(self, value):
        """Setter to Diameter Header Command Flags field.
        """
        if isinstance(value, int) and (value >= 0 and value <= 255):
            self._flags = convert_to_1_byte(value)

        elif isinstance(value, bytes) and len(value) == 1:
            self._flags = value

        elif isinstance(value, int) and (value < 0 or value > 255):
            raise DiameterHeaderAttributeValueError("flags attribute has "\
                                                    "1-byte length long")

        elif isinstance(value, bytes) and len(value) != 1:
            raise DiameterHeaderAttributeValueError("flags attribute has "\
                                                    "1-byte length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid flags "\
                                                    "attribute value")


    @property
    def command_code(self):
        """Getter to Diameter Header Command Code field.
        """
        return self._command_code


    @command_code.setter
    def command_code(self, value):
        """Setter to Diameter Header Command Code field.
        """
        if value is None:
            self._command_code = value

        elif isinstance(value, int) and (value >= 0 and value <= 16777215):
            self._command_code = convert_to_3_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 16777215):
            raise DiameterHeaderAttributeValueError("command_code attribute "\
                                                    "has 3-bytes length long")

        elif isinstance(value, bytes) and len(value) == 3:
            self._command_code = value

        elif isinstance(value, bytes) and len(value) != 3:
            raise DiameterHeaderAttributeValueError("command_code attribute "\
                                                    "has 3-bytes length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid command_code "\
                                                    "attribute value")


    @property
    def application_id(self):
        """Getter to Diameter Header Application-ID field.
        """
        return self._application_id


    @application_id.setter
    def application_id(self, value):
        """Setter to Diameter Header Application-ID field.
        """
        if value is None:
            self._application_id = value

        elif isinstance(value, int) and (value >= 0 and value <= 4294967295):
            self._application_id = convert_to_4_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 4294967295):
            raise DiameterHeaderAttributeValueError("application_id attribute "\
                                                    "has 4-bytes length long")

        elif isinstance(value, bytes) and len(value) == 4:
            self._application_id = value

        elif isinstance(value, bytes) and len(value) != 4:
            raise DiameterHeaderAttributeValueError("application_id attribute "\
                                                    "has 4-bytes length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid application_id "\
                                                    "attribute value")


    @property
    def hop_by_hop(self):
        """Getter to Diameter Header Hop-by-Hop Identifier field.
        """
        return self._hop_by_hop


    @hop_by_hop.setter
    def hop_by_hop(self, value):
        """Setter to Diameter Header Hop-by-Hop Identifier field.
        """
        if value is None:
            self._hop_by_hop = value

        elif isinstance(value, int) and (value >= 0 and value <= 4294967295):
            self._hop_by_hop = convert_to_4_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 4294967295):
            raise DiameterHeaderAttributeValueError("hop_by_hop attribute "\
                                                    "has 4-bytes length long")

        elif isinstance(value, bytes) and len(value) == 4:
            self._hop_by_hop = value

        elif isinstance(value, bytes) and len(value) != 4:
            raise DiameterHeaderAttributeValueError("hop_by_hop attribute "\
                                                    "has 4-bytes length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid hop_by_hop "\
                                                    "attribute value")


    @property
    def end_to_end(self):
        """Getter to Diameter Header End-to-End Identifier field.
        """
        return self._end_to_end


    @end_to_end.setter
    def end_to_end(self, value):
        """Setter to Diameter Header End-to-End Identifier field.
        """
        if value is None:
            self._end_to_end = value

        elif isinstance(value, int) and (value >= 0 and value <= 4294967295):
            self._end_to_end = convert_to_4_bytes(value)

        elif isinstance(value, int) and (value < 0 or value > 4294967295):
            raise DiameterHeaderAttributeValueError("end_to_end attribute "\
                                                    "has 4-bytes length long")

        elif isinstance(value, bytes) and len(value) == 4:
            self._end_to_end = value

        elif isinstance(value, bytes) and len(value) != 4:
            raise DiameterHeaderAttributeValueError("end_to_end attribute "\
                                                    "has 4-bytes length long")

        elif not isinstance(value, bytes) and not isinstance(value, int):
            raise DiameterHeaderAttributeValueError("invalid end_to_end "\
                                                    "attribute value")


    def set_request_bit(self, state):
        """Set / unset the request bit in Command Flags bit.
        """
        if not isinstance(state, bool):
            raise DiameterHeaderError("R-bit is of type Boolean")

        if self.is_error():
            raise DiameterHeaderError("R-bit MUST NOT be set when E-bit is set")

        if state:
            if self.is_request():
                raise DiameterHeaderError("R-bit was already set")

            flags = (self.get_flags() | 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_request_bit))

        else:
            if not self.is_request():
                raise DiameterHeaderError("R-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_request_bit))
        
        self.flags = convert_to_1_byte(flags)


    def is_request(self):
        """Checks if request bit is set / unset in Command Flags bit.
        """
        return (self.get_flags() & 
                DiameterHeader.get_flags_bit(DiameterHeader.flag_request_bit) != 0)


    def set_proxiable_bit(self, state):
        """Set / unset the proxiable bit in Command Flags bit.
        """
        if not isinstance(state, bool):
            raise DiameterHeaderError("P-bit is of type Boolean")

        if state:
            if self.is_proxiable():
                raise DiameterHeaderError("P-bit was already set")

            flags = (self.get_flags() | 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_proxiable_bit))
        else:
            if not self.is_proxiable():
                raise DiameterHeaderError("P-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_proxiable_bit))

        self.flags = convert_to_1_byte(flags)


    def is_proxiable(self):
        """Checks if proxiable bit is set / unset in Command Flags bit.
        """
        return (self.get_flags() & 
                DiameterHeader.get_flags_bit(DiameterHeader.flag_proxiable_bit) != 0)


    def set_error_bit(self, state):
        """Set / unset the error bit in Command Flags bit.
        """
        if not isinstance(state, bool):
            raise DiameterHeaderError("E-bit is of type Boolean")

        if self.is_request():
            raise DiameterHeaderError("E-bit MUST NOT be set when R-bit is set")

        if state:
            if self.is_error():
                raise DiameterHeaderError("E-bit was already set")

            flags = (self.get_flags() | 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_error_bit))

        else:
            if not self.is_error():
                raise DiameterHeaderError("E-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_error_bit))

        self.flags = convert_to_1_byte(flags)


    def is_error(self):
        """Checks if error bit is set / unset in Command Flags bit.
        """
        return (self.get_flags() & 
                DiameterHeader.get_flags_bit(DiameterHeader.flag_error_bit) != 0)


    def set_retransmitted_bit(self, state):
        """Set / unset the retransmitted bit in Command Flags bit.
        """
        if not isinstance(state, bool):
            raise DiameterHeaderError("T-bit is of type Boolean")

        if state:
            if self.is_retransmitted():
                raise DiameterHeaderError("T-bit was already set")

            flags = (self.get_flags() | 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_retransmitted))

        else:
            if not self.is_retransmitted():
                raise DiameterHeaderError("T-bit was already unset")

            flags = (self.get_flags() ^ 
                     DiameterHeader.get_flags_bit(DiameterHeader.flag_retransmitted))

        self.flags = convert_to_1_byte(flags)


    def is_retransmitted(self):
        """Checks if retransmitted bit is set / unset in Command Flags bit.
        """
        return (self.get_flags() & 
                DiameterHeader.get_flags_bit(DiameterHeader.flag_retransmitted) != 0)


    def get_version(self):
        """Returns the Diameter Header Version field value in Integer format.
        """
        return int.from_bytes(self.version, byteorder="big")


    def get_length(self):
        """Returns the Diameter Header Length field value in Integer format.
        """
        return int.from_bytes(self.length, byteorder="big")


    def get_flags(self):
        """Returns the Diameter Header Command Flags field value in Integer 
        format.
        """
        return int.from_bytes(self.flags, byteorder="big")


    def get_command_code(self):
        """Returns the Diameter Header Command Code field value in Integer 
        format.
        """
        return int.from_bytes(self.command_code, byteorder="big")


    def get_application_id(self):
        """Returns the Diameter Header Application-ID field value in Integer 
        format.
        """
        return int.from_bytes(self.application_id, byteorder="big")


    def get_hop_by_hop(self):
        """Returns the Diameter Header Hop-by-Hop Identifier field value in 
        Integer format.
        """
        return int.from_bytes(self.hop_by_hop, byteorder="big")


    def get_end_to_end(self):
        """Returns the Diameter Header End-to-End Identifier field value in 
        Integer format.
        """
        return int.from_bytes(self.end_to_end, byteorder="big")


    @staticmethod
    def get_flags_bit(flag_bit):
        """Returns the Diameter Header Command Flags bit in Integer format of a 
        given flag bit in hexadecimal value.
        """
        return int.from_bytes(flag_bit, byteorder="big")


    def dump(self):
        """Dump a byte stream which represents a DiameterHeader object.
        """
        dump = self.version + self.length + self.flags
        if self.command_code:
            dump += self.command_code
        if self.application_id:
            dump += self.application_id
        if self.hop_by_hop:
            dump += self.hop_by_hop
        if self.end_to_end:
            dump += self.end_to_end
        
        return dump


    @classmethod
    def load(cls, stream):
        """Load a byte stream which represents Diameter Headers and returns a 
        list of DiameterHeader objects.
        """
        version  = convert_to_1_byte(stream[0])
        length = stream[1:4]        
        flags = convert_to_1_byte(stream[4])
        command_code = stream[5:8]
        application_id = stream[8:12]
        hop_by_hop = stream[12:16]
        end_to_end = stream[16:20]

        return cls(version=version, 
                   length=length, 
                   flags=flags, 
                   command_code=command_code, 
                   application_id=application_id,
                   hop_by_hop=hop_by_hop,
                   end_to_end=end_to_end)

    
class DiameterMessage:
    """Implementation of a Diameter Message. 
    
    Refer to Section 3 of IETF RFC 6733 for details. The constructor has
    input arguments represented by DiameterHeader object and a list of 
    DiameterAVP object. Together they define the set of Diameter Message 
    attributes.

    Provides a general-case interface for create custom and standard Diameter 
    Message.

    :param header: the DiameterHeader object which represents the Diameter 
        Header format within a Diameter Message.
    :param avps: the list of DiameterAVP objects, which represents the set of
        available AVPs within a Diameter Message.

    Usage::

        >>> from bromelia.base import DiameterMessage
        >>> message = DiameterMessage()
        >>> message
        <Diameter Message: Unknown [], 0 [Diameter common message], 0 AVP(s)>
    """

    def __init__(self, header=None, avps=None, loaded=False):
        self.header = header
        self.avps = list()
        self._loaded = loaded
        
        if avps is not None:
            if not isinstance(avps, list):
                raise DiameterMessageError("invalid input argument: 'avps'. "\
                                           "It MUST be a list of DiameterAVP "\
                                           "objects")

            for idx, avp in enumerate(avps):
                if not isinstance(avp, DiameterAVP):
                    raise DiameterMessageError(f"invalid element found in "\
                                               f"list argument: 'avps'. The "\
                                               f"element {avp} in position "\
                                               f"{idx} does not represent a "\
                                               f"DiameterAVP object")

                self.append(avp)
            self._loaded = False


    def __repr__(self):
        representations = header_representation(self.header)
        representations.update({"num_of_avps": len(self.avps)})
    
        return "<Diameter Message: {cmd_code_int} [{cmd_code_str}]"\
                "{flag_representation}, {app_id_int} [{app_id_str}], "\
                "{num_of_avps} AVP(s)>".format(**representations)


    def __add__(self, other):
        """Dunder method to concatenate two DiameterMessage objects and return 
        a byte stream representing those two DiameterMessage objects.
        """
        return self.dump() + other.dump()


    def __len__(self):
        """Dunder method to get the length of a DiameterMessage object as per 
        the Diameter Message Length field.
        """
        return self.get_length()


    def __eq__(self, other):
        """Dunder method to check if the byte stream of two DiameterMessage 
        objects are equal.
        """
        return self.dump() == other.dump()


    def __bytes__(self):
        """Dunder method to return the byte stream of a DiameterMessage object. 
        It is the dump() method used in another way.
        """
        return self.dump()


    def copy(self):
        """Creates a deepcopy of a DiameterMessage object.
        """
        return deepcopy(self)


    @classmethod
    def convert(cls, msg):
        """Used to convert a specialized DiameterMessage object into a generic 
        DiameterMessage object.
        """
        if not isinstance(msg, DiameterMessage):
            raise DiameterMessageError("invalid message. It MUST be a "\
                                       "DiameterMessage subclass object to be "\
                                       "converted into DiameterMessage object")
        
        return cls(header=msg.header,
                   avps=msg.avps)


    def append(self, avp):
        """Appends a DiameterAVP object into a DiameterMessage object. It 
        creates a new brand DiameterMessage attribute with the name of the 
        DiameterAVP object. It updates the Diameter Message Length field as 
        per the new Diameter AVP length.
        """
        if not isinstance(avp, DiameterAVP):
            raise DiameterMessageError(f"cannot append a data type of "\
                                       f"'{type(avp)}'")

        #: Get the AVP class name by calling the DiameterAvpLoader object. In
        #: case its helper method does not find any reference (in other words,
        #: it returns "Unknown"), it looks by calling another helper method as 
        #: a fallback procedure.
        avp_name = loader.get_avp_class_name(avp)
        if avp_name == "Unknown":
            avp_name = avp_look_up(avp)

        #: DiameterMessage attributes must follow the Diameter AVP name in 
        #: lower, separated by underscores and suffixed with a "avp" string.
        _name = avp_name.replace("-", "_").lower()
        avp_key = f"{_name}_avp"

        #: Sometimes a DiameterMessage object may have multiples DiameterAVP
        #: objects of the same type. It appends an index at the end of the 
        #: DiameterMessage attribute name in order to not overwritting the 
        #: previous one. 
        if avp_key in self.__dict__:
            index = 0
            for key in self.__dict__.keys():
                if avp_key in key:
                    index += 1
            avp_key = f"{avp_key}__{index}"

        #: Updates DiameterMessage attributes.
        self._avps.append(avp)
        self.__dict__.update({avp_key: avp})

        #: In case this DiameterMessage object was not created by calling its
        #: load method, it updates the DiameterMessage object length attribute 
        #: with the DiameterAVP object length. 
        if not self._loaded:
            header_length = self.header.get_length() + avp.get_length()

            if avp.get_padding_length():
                header_length += avp.get_padding_length()
            self.header.length = convert_to_3_bytes(header_length)


    def extend(self, avps):
        """Extends the DiameterMessage object by appending several DiameterAVP 
        objects defined in a Python list.
        """
        for avp in avps:
            self.append(avp)


    def has_avp(self, avp_key):
        """Checks if a given Diameter AVP is defined in the DiameterMessage 
        object. Once DiameterMessage objects have attributes with the 
        DiameterAVP objects name, it checks the key inside the __dict__ 
        attribute.
        """
        if not isinstance(avp_key, str):
            raise DiameterMessageError("`avp_key` must be str")

        if not self.avps:
            return False
        elif avp_key in self.__dict__:
            return True
        return False


    def pop(self, avp_key):
        """Remove a DiameterAVP object from a DiameterMessage object based on
        Diameter AVP name.
        """
        if not self.avps:
            raise DiameterMessageError("`avps` attribute is empty. There is "\
                                       "no DiameterAVP object to be removed")

        avp = self.__dict__[avp_key]

        #: Updates DiameterMessage attributes.
        self._avps.remove(avp)
        self.__dict__.pop(avp_key, None)

        #: It updates the DiameterMessage object length attribute with the 
        #: DiameterAVP object length.
        header_length = self.header.get_length() - avp.get_length()
        if avp.get_padding_length():
            header_length -= avp.get_padding_length()
        self.header.length = convert_to_3_bytes(header_length)


    def update_key(self, old_avp_key, new_avp_key):
        """Updates the DiameterMessage attribute which refer to a given 
        DiameterAVP object.
        """
        if not self.has_avp(old_avp_key):
            raise DiameterMessageError(f"`{old_avp_key}` key not defined")
            
        if self.has_avp(new_avp_key):
            raise DiameterMessageError(f"`{new_avp_key}` key already defined")

        self.__dict__[new_avp_key] = self.__dict__.pop(old_avp_key)


    def cleanup(self):
        """Cleanup all the DiameterMessage attributes and its respective 
        DiameterAVP objects.
        """
        self._avps = list()

        #: Gets all DiameterMessage attributes based on DiameterAVP objects.
        avps_keys = list()
        for avp_key in self.__dict__.keys():
            if "_avp" in avp_key and avp_key != "_avps":
                avps_keys.append(avp_key)

        #: Goes over each DiameterMessage attribute based on DiameterAVP 
        #: object, pops it up and updates the DiameterMessage length attribute
        #: by decreasing the length as per the DiameterAVP length.
        for avp_key in avps_keys:
            item = self.__dict__[avp_key]

            self.__dict__.pop(avp_key, None)

            header_length = self.header.get_length() - item.get_length()
            if item.get_padding_length():
                header_length -= item.get_padding_length()
            self.header.length = convert_to_3_bytes(header_length)


    @staticmethod
    def load(stream):
        """Load a byte stream which represents Diameter Message and returns a 
        list of DiameterMessage objects.
        """
        messages = []
        index = 0

        while index < len(stream):
            header_stream = stream[index:index+DIAMETER_HEADER_LENGTH]
            header = DiameterHeader.load(header_stream)

            lower_limit = index + DIAMETER_HEADER_LENGTH
            upper_limit = index + header.get_length()

            avp_stream = stream[lower_limit:upper_limit]
            avps = DiameterAVP.load(avp_stream)

            message = DiameterMessage(header, avps, loaded=True)
            messages.append(message)

            index += header.get_length()

        return messages


    def _load(self, values):
        _kwargs = values.get("kwargs", None)
        if _kwargs:
            values.update(_kwargs)

        values.pop("kwargs", None)
        values.pop("self", None)

        avps_info = values.items()

        kwargs = dict()
        for avp_name, avp_value in avps_info:
            _name = f"{avp_name}_avp"

            if avp_name in self.mandatory and avp_value is None:
                raise DiameterMessageError(f"missing mandatory AVP "\
                                           f"argument avp_value: '{avp_name}'")

            elif avp_name in self.mandatory and avp_value is not None:
                _value = self.mandatory[avp_name](avp_value)
                self.append(_value)
            
            elif avp_name in self.optionals and avp_value is not None:
                _value = self.optionals[avp_name](avp_value)
                self.append(_value)

            elif avp_name not in self.optionals and avp_value is not None:
                if not isinstance(avp_value, DiameterAVP):
                    raise DiameterMessageError("non-mandatory and "\
                                               "non-optionals AVPs should "\
                                               "include DiameterAVP object "\
                                               "only")
                self.append(avp_value)

            self.__dict__.update(kwargs)

        self.header.length = self.length


    @property
    def header(self):
        """Getter to header attribute.
        """
        return self._header


    @header.setter
    def header(self, value):
        """Setter to header attribute.
        """
        if value is None:
            self._header = DiameterHeader()
        else:
            self._header = value


    @property
    def avps(self):
        """Getter to avps attribute.
        """
        return tuple(self._avps)


    @avps.setter
    def avps(self, value):
        """Setter to avps attribute. It calls either the append or the extend
        methods depending on the number of DiameterAVP objects inside the list. 
        """
        if not isinstance(value, list):
            raise DiameterMessageError(f"only list allowed. Cannot append a "\
                                       f"data type of '{type(value)}'")
    
        self.cleanup()

        if len(value) == 1:
            self.append(value[0])
        elif len(value) > 1:
            self.extend(value)


    def __getitem__(self, idx):
        return self._avps[idx]


    def __setitem__(self, idx, value):
        self._avps[idx] = value


    @property
    def loaded(self):
        """Getter to loaded attribute.
        """
        return self._loaded


    @loaded.setter
    def loaded(self, value):
        """Setter to loaded attribute.
        """
        if not isinstance(value, bool):
            raise DiameterMessageError("only bool allowed")
                
        self._loaded = value            


    #: TO-DO: review it
    def refresh(self):
        real_length = 20
        for avp in self.avps:
            real_length += len(avp)
            padding = avp.get_padding_length()
            if padding:
                real_length += avp.get_padding_length()

        if real_length != self.header.get_length():
            self.header.length = convert_to_3_bytes(real_length)

        
    @property
    def length(self):
        self.refresh()
        return self.header.length


    def get_version(self):
        """Returns the Diameter Header Version field value in Integer format.
        """
        return int.from_bytes(self.header.version, byteorder="big")


    def get_length(self):
        """Returns the Diameter Header Length field value in Integer format.
        """
        return int.from_bytes(self.length, byteorder="big")


    def get_flags(self):
        """Returns the Diameter Header Command Flags field value in Integer 
        format.
        """
        return int.from_bytes(self.header.flags, byteorder="big")


    def get_command_code(self):
        """Returns the Diameter Header Command Code field value in Integer 
        format.
        """
        return int.from_bytes(self.header.command_code, byteorder="big")


    def get_application_id(self):
        """Returns the Diameter Header Application-ID field value in Integer 
        format.
        """
        return int.from_bytes(self.header.application_id, byteorder="big")


    def get_hop_by_hop(self):
        """Returns the Diameter Header Hop-by-Hop Identifier field value in 
        Integer format.
        """
        return int.from_bytes(self.header.hop_by_hop, byteorder="big")


    def get_end_to_end(self):
        """Returns the Diameter Header End-to-End Identifier field value in 
        Integer format.
        """
        return int.from_bytes(self.header.end_to_end, byteorder="big")
    

    def set_flag_by_app_id(self, app_id):
        """Set / unset the Command Flags bit as per Application-ID field value.
        """
        if app_id == DIAMETER_APPLICATION_DEFAULT:
            if self.header.is_proxiable():
                self.header.set_proxiable_bit(False)
        else:
            self.header.set_proxiable_bit(True)

        if isinstance(self, DiameterRequest):
            self.header.set_request_bit(True)

        elif isinstance(self, DiameterAnswer):
            if self.header.is_request():
                self.header.set_request_bit(False)


    def dump(self):
        """Dump a byte stream which represents a DiameterMessage object.
        """
        dump = self.header.dump()
        for avp in self.avps:
            dump += avp.dump()
        return dump


class DiameterRequest(DiameterMessage):
    """Implementation of a Diameter Request.
    
    Refer to Section 3 of IETF RFC 6733 for details. It relies on the 
    DiameterMessage. The difference is that DiameterRequest set 'R-bit' and 
    implements two private methods to calculate the Hop-by-Hop and End-to-End
    fields.

    :param application_id: represents the Application-ID field.
    :param command_code: represents the Command Code field.

    Usage::

        >>> from bromelia.base import DiameterRequest
        >>> request = DiameterRequest()
        >>> request
        <Diameter Message: Unknown [] REQ, 0 [Diameter common message], 0 AVP(s)>
    """

    hop_by_hop_identifiers = list()
    end_to_end_identifiers = list()

    def __init__(self, 
                 version=DIAMETER_VERSION, 
                 command_code=0, 
                 application_id=0,
                 header=None,
                 avps=None):

        if header:
            _header = DiameterHeader(version=header.version,
                                    command_code=header.command_code,
                                    application_id=header.application_id,
                                    hop_by_hop=header.hop_by_hop,
                                    end_to_end=header.end_to_end)
        else:
            hop_by_hop = self.__set_hop_by_hop_identifier()
            end_to_end = self.__set_end_to_end_identifier()

            _header = DiameterHeader(version=version,
                                    command_code=command_code,
                                    application_id=application_id,
                                    hop_by_hop=hop_by_hop,
                                    end_to_end=end_to_end)

        DiameterMessage.__init__(self, _header, avps)
        DiameterMessage.set_flag_by_app_id(self, application_id)        


    def __set_hop_by_hop_identifier(self):
        """Sets the Hop-by-Hop Identifier field in Diameter Header"""
        while True:
            random_identifier = os.urandom(4)
            if random_identifier not in DiameterRequest.hop_by_hop_identifiers:
                DiameterRequest.hop_by_hop_identifiers.append(random_identifier)
                return random_identifier


    def __set_end_to_end_identifier(self):
        """Sets the End-to-End Identifier field in Diameter Header"""
        while True:
            random_identifier = os.urandom(4)
            if random_identifier not in DiameterRequest.end_to_end_identifiers:
                DiameterRequest.end_to_end_identifiers.append(random_identifier)
                return random_identifier


    @staticmethod
    def convert(msg):
        """Method not implemented to be used in the specialized DiameterRequest 
        object. It should be only used to convert more specialized 
        DiameterMessage objects.
        """
        raise DiameterMessageError("DiameterRequest class does not have "\
                                   "this method available for use")


class DiameterAnswer(DiameterMessage):
    """Implementation of a Diameter Answer.
    
    Refer to Section 3 of IETF RFC 6733 for details. It relies on the 
    DiameterMessage. The difference is that DiameterAnswer does not set 
    'R-bit'.

    :param application_id: represents the Application-ID field.
    :param command_code: represents the Command Code field.

    Usage::

        >>> from bromelia.base import DiameterAnswer
        >>> answer = DiameterAnswer()
        >>> answer
        <Diameter Message: Unknown [], 0 [Diameter common message], 0 AVP(s)>
    """

    def __init__(self, 
                 version=DIAMETER_VERSION, 
                 command_code=0, 
                 application_id=0,
                 header=None,
                 avps=None):

        if header:
            _header = DiameterHeader(version=header.version,
                                     command_code=header.command_code,
                                     application_id=header.application_id,
                                     hop_by_hop=header.hop_by_hop,
                                     end_to_end=header.end_to_end)
        else:
            _header = DiameterHeader(version=version,
                                     command_code=command_code, 
                                     application_id=application_id)

        DiameterMessage.__init__(self, _header, avps)
        DiameterMessage.set_flag_by_app_id(self, application_id)


    @staticmethod
    def convert(msg):
        """Method not implemented to be used in the specialized DiameterRequest 
        object. It should be only used to convert more specialized 
        DiameterMessage objects.
        """
        raise DiameterMessageError("DiameterAnswer class does not have "\
                                "this method available for use")


loader = DiameterAvpLoader()



# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_swm.definitions
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Defines handy structures that are used within the library for the 
    3GPP SWm Application Id.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""


class EapPayload(object):
    """Implementation of EAP Protocol as per IETF RFC 3748.
    """

    __slots__ = ("eap_code", "eap_id", "length", "eap_type", "content")

    def __init__(self, eap_code, eap_id, eap_type, content):          
        self.eap_code = eap_code
        self.eap_id = eap_id
        self.length = 1
        self.eap_type = eap_type
        self.content = content



# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_s13.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol messages for 3GPP S13/S13'
    Application Id.

    :copyright: (c) 2021-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from .avps import *

from ...base import DiameterRequest, DiameterAnswer
from ...constants import *


class MeIdentityCheckAnswer(DiameterAnswer):
    """Implementation of ME-Identity-Check-Answer (ECA) command as per 
    clause 7.2.20 of ETSI TS 129 272 V15.4.0 (2018-07).

    The ME-Identity-Check-Answer is indicated by the Command Code field 
    set to 324 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s13 import ECA
        >>> eca = ECA()
        >>> eca
        <Diameter Message: 324 [ECA] PXY, 16777252 [3GPP S13], 6 AVP(s)>    
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = { 
                    #"drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "equipment_status": EquipmentStatusAVP,
                    "failed_avp": FailedAvpAVP,
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S13_S13)],
                 result_code=DIAMETER_SUCCESS,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED, 
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 equipment_status=None,
                 avp=None,
                 failed_avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=EC_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S13_S13)

        DiameterAnswer._load(self, locals())


class MeIdentityCheckRequest(DiameterRequest):
    """Implementation of ME-Identity-Check-Request (ECR) command as per 
    clause 7.2.19 of ETSI TS 129 272 V15.4.0 (2018-07).

    The ME-Identity-Check-Request is indicated by the Command Code field 
    set to 324 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s13 import ECR
        >>> ecr_avps = {
        ...     "destination_realm": "remote",
        ...     "terminal_information": [ImeiAVP("12345678901234"), SoftwareVersionAVP("12")],
        ... }
        >>> ecr = ECR(**ecr_avps)
        >>> ecr
        <Diameter Message: 324 [ECR] REQ|PXY, 16777252 [3GPP S13], 8 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "terminal_information": TerminalInformationAVP,
    }
    optionals = {
                    #"drmp": DrmpAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "destination_host": DestinationHostAVP,
                    "user_name": UserNameAVP, 
                    "proxy_info": ProxyInfoAVP,
                    "route_record": RouteRecordAVP,
    }

    def __init__(self,
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_S13_S13)],
                 auth_session_state=NO_STATE_MAINTAINED, 
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 terminal_information=None,
                 user_name=None,
                 avp=None,
                 proxy_info=None,
                 route_record=None,
                 **kwargs):

        DiameterRequest.__init__(self,
                                 command_code=EC_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S13_S13)

        DiameterRequest._load(self, locals())


# -*- coding: utf-8 -*-
"""
    bromelia.setup
    ~~~~~~~~~~~~~~
    
    This module implements the central Diameter application object. It works
    as per Peer State Machine defined in RFC 6733 in order to establish a 
    Diameter association with another Peer Node.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import copy
import datetime
import logging
import platform
import queue
import socket 
import sys
import threading
import time
from contextlib import contextmanager

from ._internal_utils import _convert_config_to_connection_obj
from ._internal_utils import get_app_ids
from ._internal_utils import application_id_look_up
from .base import DiameterMessage
from .config import Config
from .config import DiameterLogging
from .config import LISTENING_TICKER
from .config import SEND_BUFFER_MAXIMUM_SIZE
from .config import SLEEP_TIMER
from .config import WAITING_CONN_TIMER
from .constants import DIAMETER_AGENT_CLIENT_MODE
from .constants import DIAMETER_AGENT_SERVER_MODE
from .exceptions import AVPParsingError
from .exceptions import DiameterApplicationError
from .exceptions import DiameterAssociationError
from .messages import DiameterAnswer
from .messages import DiameterRequest
from .proxy import DiameterBaseProxy
from .statemachine import PeerStateMachine
from .statemachine import CLOSED, I_OPEN, R_OPEN
from .transport import TcpClient
from .transport import TcpServer
from .utils import is_base_request
from .utils import is_base_answer

diameter_conn_logger = logging.getLogger("DiameterConnection")
diameter_logger = logging.getLogger("Diameter")


class DiameterAssociation(object):
    def __init__(self, connection, base):
        self.connection = connection
        self.base = base

        self.state_is_active = False
        self.transport = None
        self.error_has_raised = False
        self._stop_threads = False

        self.num_answers = 0
        self.num_requests = 0

        self.watchdog_timeout = self.connection.watchdog_timeout
        self.tracking_events_count = 0

        self.end_to_end_identifiers = list()
        self.pending_requests = dict()

        self._recv_messages = queue.Queue()
        self._send_messages = queue.Queue()

        self.postprocess_recv_messages = queue.Queue() 
        self.postprocess_recv_messages_ready = threading.Event()
        self.postprocess_recv_messages_lock = threading.Lock()
        self.lock = threading.Lock()


    def is_connected(self):
        if self.transport:
           return self.transport.is_connected
        
        return False


    def __is_connected(self):
        if not self.is_connected():
            raise DiameterAssociationError("There is no transport "\
                                           "connection up for this PeerNode.")


    def start(self):
        self._stop_threads = False

        if self.connection.mode == DIAMETER_AGENT_CLIENT_MODE:
               self.transport = TcpClient(self.connection.peer_node.ip_address, 
                                          self.connection.peer_node.port)
            
        elif self.connection.mode == DIAMETER_AGENT_SERVER_MODE:
            self.transport = TcpServer(self.connection.local_node.ip_address, 
                                       self.connection.local_node.port)

        else:
            raise DiameterAssociationError("Invalid Diameter Agent mode.")

        self.transport.start()
        self.transport.run()

        threading.Thread(name="recv_message_from_queue",
                         target=self.recv_message_from_queue).start()


    def close(self):
        self.__is_connected()

        self.state_is_active = False
        self._stop_threads = True
        self.transport.close()
        self.transport = None


    def recv_message_from_queue(self):
        while not self._stop_threads and self.transport:
            self.transport._recv_data_available.wait()

            self.lock.acquire()

            data_stream = copy.copy(self.transport._recv_data_stream)
            self.transport._recv_data_stream = b""
            self.transport._recv_data_available.clear()

            diameter_conn_logger.debug("Grabbing data stream from "\
                                       "Transport Layer to Diameter Layer.")

            try:
                msgs = DiameterMessage.load(data_stream)
                for msg in msgs:
                    if msg.has_avp("user_name_avp"):
                        diameter_conn_logger.debug(f"Message from Transport Layer: "\
                                                   f"{application_id_look_up(msg.header.application_id)[0]}, "\
                                                   f"{msg.header.get_command_code()}, "\
                                                   f"{msg.header.hop_by_hop.hex()}, "\
                                                   f"LEN: {len(msgs)}, "\
                                                   f"DATA_LENGTH: {len(data_stream)}, "\
                                                   f"USERNAME: {msg.user_name_avp.data}")
                    self._recv_messages.put(msg)
                
                diameter_conn_logger.debug(f"Found {len(msgs)} Diameter "\
                                           f"Message(s).")
            except AVPParsingError:
                diameter_conn_logger.exception(f"AVPParsingError has "\
                                               f"been raised due stream: "\
                                               f"{self.transport._recv_data_stream.hex()}")

            self.lock.release()


    def put_message_into_send_queue(self, msg):
        self.lock.acquire()

        self.__is_connected()
        self._send_messages.put(msg)

        hop_by_hop = msg.header.hop_by_hop

        if isinstance(msg, DiameterRequest):
            diameter_conn_logger.debug(f"[{hop_by_hop.hex()}] Diameter "\
                                       f"Request have been put into "\
                                       f"_send_messages Queue.")

            key = msg.header.end_to_end.hex()
            self.end_to_end_identifiers.append(key)

        elif isinstance(msg, DiameterAnswer):
            diameter_conn_logger.debug(f"[{hop_by_hop.hex()}] Diameter "\
                                       f"Answer have been put into "\
                                       f"_send_messages Queue.")

        elif isinstance(msg, DiameterMessage):
            if msg.header.is_request():
                diameter_conn_logger.debug(f"[{hop_by_hop.hex()}] Diameter "\
                                           f"Message (Request) have been put "\
                                           f"into _send_messages Queue.")

                key = msg.header.end_to_end.hex()
                self.end_to_end_identifiers.append(key)

            else:
                diameter_conn_logger.debug(f"[{hop_by_hop.hex()}] Diameter "\
                                           f"Message (Answer) have been put "\
                                           f"into _send_messages Queue.")
                
        self.lock.release()


    def send_message_from_queue(self):
        self.lock.acquire()
        self.__is_connected()

        diameter_conn_logger.debug(f"There is/are "\
                                   f"{self._send_messages.qsize()} Diameter "\
                                   f"Message(s) in the Sending Queue.")

        stream = b""
        while not self._send_messages.empty() and \
                len(stream) <= SEND_BUFFER_MAXIMUM_SIZE:
            msg = self._send_messages.get()
            diameter_conn_logger.debug(f"[{msg.header.hop_by_hop.hex()}] "\
                                       f"Preparing message to be sent.")

            MESSAGE_LENGTH = len(msg.dump())

            if MESSAGE_LENGTH > SEND_BUFFER_MAXIMUM_SIZE - len(stream):
                self._send_messages.put(msg)
                break

            if isinstance(msg, DiameterRequest):
                key = msg.header.hop_by_hop.hex()
                self.pending_requests.update({key: msg})
                diameter_conn_logger.debug(f"[{msg.header.hop_by_hop.hex()}] "\
                                           f"Diameter Request have been "\
                                           f"put into Pending Request Queue.")
    
            stream += msg.dump()

        if self.transport:
            if not self.transport.is_write_mode():
                diameter_conn_logger.debug("Transport Layer is not in WRITE "\
                                           "mode, so we can send data stream.")

                self.transport._set_selector_events_mask("rw", stream)
            else:
                diameter_conn_logger.debug("Transport Layer is in WRITE "\
                                           "mode, so we cannot send data "\
                                           "stream.")

                while not self._stop_threads and self.transport:
                    self.transport.write_mode_on.wait()
                    if not self.transport.is_write_mode():
                        diameter_conn_logger.debug("Transport Layer is not in "\
                                                   "WRITE mode again, so we "\
                                                   "can send data stream.")
    
                        self.transport._set_selector_events_mask("rw", stream)
                        
                        # maybe include a verification here before the "break" if a given message has been sent from transport layer.
                        break


        self.lock.release()


    def get_message(self):
        while not self._stop_threads:
            if self.postprocess_recv_messages.empty():
                self.postprocess_recv_messages_ready.wait()
                diameter_conn_logger.debug("Got go ahead for "\
                                           "postprocess_recv_messages_ready")
            else:
                diameter_conn_logger.debug("No need to wait for go ahead for "\
                                           "postprocess_recv_messages_ready")
    
            self.lock.acquire()
            diameter_conn_logger.debug("Acquired DiameterAssociation lock")

            self.postprocess_recv_messages_lock.acquire()
            message = self.postprocess_recv_messages.get()
            self.postprocess_recv_messages_lock.release()

            if message.has_avp("user_name_avp"):
                diameter_conn_logger.debug(f"Message from "\
                                           f"postprocess_recv_messages Queue: "\
                                           f"{application_id_look_up(message.header.application_id)[0]}, "\
                                           f"{message.header.get_command_code()}, "\
                                           f"{message.header.is_request()}, "\
                                           f"{message.header.hop_by_hop.hex()}, "\
                                           f"{len(message)}, "\
                                           f"{message.user_name_avp.data}")
            else:
                diameter_conn_logger.debug(f"Message from "\
                                           f"postprocess_recv_messages Queue: "\
                                           f"{application_id_look_up(message.header.application_id)[0]}, "\
                                           f"{message.header.get_command_code()}, "\
                                           f"{message.header.is_request()}, "\
                                           f"{message.header.hop_by_hop.hex()}, "\
                                           f"{len(message)}")

            self.postprocess_recv_messages_ready.clear()
            diameter_conn_logger.debug("Cleared go ahead for "\
                                       "postprocess_recv_messages_ready")

            self.lock.release()
            diameter_conn_logger.debug("Released DiameterAssociation lock")

            return message


    def tracking_events(self):
        try:
            if (not self.transport.events) and (self.transport.tracking_events_count >= self.watchdog_timeout):
                self.put_message_into_send_queue(self.base.dwr)
                diameter_conn_logger.debug("Generating a DWR message.")

                self.transport.tracking_events_count = 0
        
        except AttributeError as e:
            if e.args[0] == "'TcpServer' object has no attribute 'events'":
                pass


class Diameter:
    config_class = Config

    default_config = {
            "MODE": "CLIENT",
            "APPLICATIONS": [],
            "LOCAL_NODE_HOSTNAME": platform.node(),
            "LOCAL_NODE_REALM": socket.getfqdn(),
            "LOCAL_NODE_IP_ADDRESS": socket.gethostbyname(platform.node()),
            "LOCAL_NODE_PORT": 3868,
            "PEER_NODE_HOSTNAME": None,
            "PEER_NODE_REALM": None,
            "PEER_NODE_IP_ADDRESS": None,
            "PEER_NODE_PORT": 3868,
            "WATCHDOG_TIMEOUT": 60
        }


    def __init__(self,
                 config=None,
                 debug=False,
                 is_logging=False,
                 app_name=None):
        
        self.logging = DiameterLogging(debug, is_logging, app_name)
        
        self.config = self.make_config(config)
        self._association = None
        self._peer_state_machine = None


    def make_config(self, config):
        if config:
            return self.config_class(config)
        return self.config_class(Diameter.default_config)


    def get_base_messages(self, msgs=None):
        proxy = DiameterBaseProxy(self._connection)
        if msgs:
            return proxy.get_custom_messages(msgs)
        else:
            return proxy.get_default_messages()


    def get_current_state(self):
        if self._peer_state_machine:
            return self._peer_state_machine.get_current_state()
        return CLOSED


    def start(self):
        current_state = self.get_current_state()

        if current_state != CLOSED:
            raise DiameterApplicationError("Cannot start the application. "\
                                           "Peer State Machine is already "\
                                           "running")

        self._connection = _convert_config_to_connection_obj(self.config)
        self._base = self.get_base_messages()

        self._association = DiameterAssociation(self._connection, self._base)
        self._peer_state_machine = PeerStateMachine(self._association)

        self._peer_state_machine.start()
        self._association.start()


    def reset(self):
        current_state = self.get_current_state()

        if current_state == CLOSED:
            raise DiameterApplicationError("Cannot reset the application. "\
                                           "Peer State Machine is already "\
                                           "closed")

        self.close()
        time.sleep(SLEEP_TIMER)
        self.start()


    def close(self):
        current_state = self.get_current_state()
        
        if current_state == CLOSED:
            raise DiameterApplicationError("Cannot stop the application. "\
                                           "Peer State Machine is already "\
                                           "closed")

        self._peer_state_machine.close()


    def send_messages(self, msgs):
        for msg in msgs:
            self._association.put_message_into_send_queue(msg)


    def send_message(self, msg, avoid=True):    
        if isinstance(msg, DiameterRequest):
            diameter_logger.debug(f"External app wants to send a Diameter "\
                                  f"Request: "\
                                  f"{application_id_look_up(msg.header.application_id)[0]}, "\
                                  f"{msg.header.get_command_code()}, "\
                                  f"{msg.header.hop_by_hop.hex()}, "\
                                  f"LEN: {len(msg)}")

            if is_base_request(msg):
                raise DiameterApplicationError("Cannot send a Base protocol "\
                                               "request")

            self._association.put_message_into_send_queue(msg)
            if not avoid:
                return self._association.get_message(msg)

        elif isinstance(msg, DiameterAnswer):
            diameter_logger.debug(f"External app wants to send a Diameter "\
                                  f"Answer: "\
                                  f"{application_id_look_up(msg.header.application_id)[0]}, "\
                                  f"{msg.header.get_command_code()}, "\
                                  f"{msg.header.hop_by_hop.hex()}, "\
                                  f"LEN: {len(msg)}")

            if is_base_answer(msg):
                raise DiameterApplicationError("Cannot send a Base protocol "\
                                               "answer")

            self._association.put_message_into_send_queue(msg)

        elif isinstance(msg, DiameterMessage):
            if msg.header.is_request():
                if is_base_request(msg):
                    raise DiameterApplicationError("Cannot send a Base protocol "\
                                                   "request")

                self._association.put_message_into_send_queue(msg)
                if not avoid:
                    return self._association.get_answer_from_request(msg)

            else:
                if is_base_answer(msg):
                    raise DiameterApplicationError("Cannot send a Base protocol "\
                                                   "answer")
                
                self._association.put_message_into_send_queue(msg)


        else:
            raise DiameterApplicationError("Either Diameter Request or "\
                                           "Diameter Answer objects are "\
                                           "allowed to be sent")


    def get_message(self):
        return self._association.get_message()


    @contextmanager
    def context(self):
        app_ids = get_app_ids(self.config["APPLICATIONS"])
        print(f"  * Running Diameter app ({app_ids}) on "\
              f"{self.config['LOCAL_NODE_IP_ADDRESS']}"\
              f":{self.config['LOCAL_NODE_PORT']} as "\
              f"{self.config['MODE']} mode (CEX)")

        try:
            self.start()

            start = datetime.datetime.utcnow()
            while True:
                time.sleep(LISTENING_TICKER)
                stop = datetime.datetime.utcnow()

                if self.is_open():
                    break
                
                if self.is_closed() and (stop - start).seconds >= 5:
                    start = datetime.datetime.utcnow()
                    self.start()

            print(f"  * Diameter connection on "\
                  f"{self.config['LOCAL_NODE_IP_ADDRESS']}:"\
                  f"{self.config['LOCAL_NODE_PORT']} is now up")

            time.sleep(WAITING_CONN_TIMER)
            yield self

            if self.is_open():
                self.close()

        except KeyboardInterrupt:
            sys.exit(0)

        print(f"  * Diameter connection on "\
              f"{self.config['LOCAL_NODE_IP_ADDRESS']}:"\
              f"{self.config['LOCAL_NODE_PORT']} is now down")


    def is_open(self):
        try:
            if self.get_current_state() == I_OPEN or \
               self.get_current_state() == R_OPEN:
                return True
            return False
        except TypeError:
            if "'NoneType' object is not subscriptable":
                return False


    def is_closed(self):
        try:
            if self.get_current_state() == CLOSED:
                return True
            return False
        except TypeError:
            if "'NoneType' object is not subscriptable":
                return False



# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf.rfc4006
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in IETF RFC 4006.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants.ietf.rfc4006 import *
from ...types import *


class SubscriptionIdDataAVP(DiameterAVP, UTF8StringType):
    """Implementation of Subscription-Id-Data AVP in Section 8.48 of 
    IETF RFC 4006.

    The Subscription-Id-Data AVP (AVP Code 444) is of type UTF8String.
    """
    code = SUBSCRIPTION_ID_DATA_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, SubscriptionIdDataAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)


class SubscriptionIdTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Subscription-Id-Type AVP in Section 8.47 of 
    IETF RFC 4006.

    The Subscription-Id-Type AVP (AVP Code 450) is of type Enumerated.
    """
    code = SUBSCRIPTION_ID_TYPE_AVP_CODE
    vendor_id = None

    values = [
                END_USER_E164,
                END_USER_IMSI,
                END_USER_SIP_URI,
                END_USER_NAI,
                END_USER_PRIVATE
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, SubscriptionIdTypeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class SubscriptionIdAVP(DiameterAVP, GroupedType):
    """Implementation of Subscription-Id AVP in Section 8.46 of IETF RFC 4006.

    The Subscription-Id AVP (AVP Code 443) is of type Grouped.
    """
    code = SUBSCRIPTION_ID_AVP_CODE
    vendor_id = None

    mandatory = {
                    "subscription_id_data": SubscriptionIdDataAVP,
                    "subscription_id_type": SubscriptionIdTypeAVP,
    }
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, SubscriptionIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_183_017
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 183 017.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants import *
from ...types import *


class ReservationPriorityAVP(DiameterAVP, EnumeratedType):
    """Implementation of Reservation-Priority AVP in Section 7.3.9 of 
    ETSI TS 183 017 V2.3.1 (2008-09).

    The Reservation-Priority AVP (AVP Code 458) is of type Enumerated.
    """
    code = RESERVATION_PRIORITY_AVP_CODE
    vendor_id = VENDOR_ID_ETSI

    values = [
                PRIORITY_DEFAULT,
                PRIORITY_ONE,
                PRIORITY_TWO,
                PRIORITY_THREE,
                PRIORITY_FOUR,
                PRIORITY_FIVE,
                PRIORITY_SIX,
                PRIORITY_SEVEN
    ]

    def __init__(self, data=PRIORITY_DEFAULT):
        DiameterAVP.__init__(self, 
                             ReservationPriorityAVP.code,
                             ReservationPriorityAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_ETSI)



# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf.rfc4072
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in IETF RFC 4072.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import base64

from ..._internal_utils import convert_to_1_byte
from ...base import DiameterAVP
from ...constants.ietf.rfc4072 import *
from ...types import *


class EapPayloadAVP(DiameterAVP, OctetStringType):
    """Implementation of EAP-Payload AVP in Section 4.1.1 of IETF RFC 4072.

    The EAP-Payload AVP (AVP Code 462) is of type OctetString.
    """
    code = EAP_PAYLOAD_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, EapPayloadAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)    

        EapPayloadAVP.parser_data(self, data)

        OctetStringType.__init__(self, data=self.data)


    def parser_data(self, data):
        if not isinstance(data, bytes):
            if data.eap_type == EAP_TYPE_IDENTITY:
                eap_nai_data = data.content["nai"].encode("utf-8")
                eap_length = convert_to_2_bytes(len(data.eap_code + convert_to_1_byte(data.eap_id) + data.eap_type + eap_nai_data) + 2)
                
                data = data.eap_code + convert_to_1_byte(data.eap_id) + eap_length + data.eap_type + eap_nai_data
                        
            elif data.eap_type == EAP_TYPE_UMTS_AUTHENTICATION_AND_KEY_AGREEMENT_EAP:
                data = base64.b64decode(data.content["payload"])

        self.data = data


class EapMasterSessionKeyAVP(DiameterAVP, OctetStringType):
    """Implementation of EAP-Master-Session-Key AVP in Section 4.1.3
    of IETF RFC 4072.

    The EAP-Master-Session-Key AVP (AVP Code 464) is of type OctetString.
    """
    code = EAP_MASTER_SESSION_KEY_AVP_CODE
    vendor_id = None

    def __init__(self, data): 
        DiameterAVP.__init__(self, EapMasterSessionKeyAVP.code)
        OctetStringType.__init__(self, data=data)


# -*- coding: utf-8 -*-
"""
    bromelia.constants.etsi_3gpp.ts_129_272
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in ETSI TS 129 272.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_3_bytes
from ..._internal_utils import convert_to_4_bytes


#: Diameter Messages
UPDATE_LOCATION_MESSAGE = convert_to_3_bytes(316)
CANCEL_LOCATION_MESSAGE = convert_to_3_bytes(317)
AUTHENTICATION_INFORMATION_MESSAGE = convert_to_3_bytes(318)
PURGE_UE_MESSAGE = convert_to_3_bytes(321)
NOTIFY_MESSAGE = convert_to_3_bytes(323)
EC_MESSAGE = convert_to_3_bytes(324)

#: Diameter AVPs
SERVICE_SELECTION_AVP_CODE = convert_to_4_bytes(493)
ALLOCATION_RETENTION_PRIORITY_AVP_CODE = convert_to_4_bytes(1034)
SUBSCRIPTION_DATA_AVP_CODE = convert_to_4_bytes(1400)
TERMINAL_INFORMATION_AVP_CODE = convert_to_4_bytes(1401)
IMEI_AVP_CODE = convert_to_4_bytes(1402)
SOFTWARE_VERSION_AVP_CODE = convert_to_4_bytes(1403)
ULR_FLAGS_AVP_CODE = convert_to_4_bytes(1405)
ULA_FLAGS_AVP_CODE = convert_to_4_bytes(1406)
VISITED_PLMN_ID_AVP_CODE = convert_to_4_bytes(1407)
REQUESTED_EUTRAN_AUTHENTICATION_INFO_AVP_CODE = convert_to_4_bytes(1408)
NUMBER_OF_REQUESTED_VECTORS_AVP_CODE = convert_to_4_bytes(1410)
RE_SYNCHRONIZATION_INFO_AVP_CODE = convert_to_4_bytes(1411)
IMMEDIATE_RESPONSE_PREFERRED_AVP_CODE = convert_to_4_bytes(1412)
AUTHENTICATION_INFO_AVP_CODE = convert_to_4_bytes(1413)
E_UTRAN_VECTOR_AVP_CODE = convert_to_4_bytes(1414)
ITEM_NUMBER_AVP_CODE = convert_to_4_bytes(1419)
CANCELLATION_TYPE_AVP_CODE = convert_to_4_bytes(1420)
CONTEXT_IDENTIFIER_AVP_CODE = convert_to_4_bytes(1423)
SUBSCRIBER_STATUS_AVP_CODE = convert_to_4_bytes(1424)
OPERATOR_DETERMINED_BARRING_AVP_CODE = convert_to_4_bytes(1425)
ALL_APN_CONFIGURATIONS_INCLUDED_INDICATOR_AVP_CODE = convert_to_4_bytes(1428)
APN_CONFIGURATION_PROFILE_AVP_CODE = convert_to_4_bytes(1429)
APN_CONFIGURATION_AVP_CODE = convert_to_4_bytes(1430)
EPS_SUBSCRIBED_QOS_PROFILE_AVP_CODE = convert_to_4_bytes(1431)
VPLMN_DYNAMIC_ADDRESS_ALLOWED_AVP_CODE = convert_to_4_bytes(1432)
STN_SR_AVP_CODE = convert_to_4_bytes(1433)
ALERT_REASON_AVP_CODE = convert_to_4_bytes(1434)
AMBR_AVP_CODE = convert_to_4_bytes(1435)
PDN_GW_ALLOCATION_TYPE_AVP_CODE = convert_to_4_bytes(1438)
PUA_FLAGS_AVP_CODE = convert_to_4_bytes(1442)
NOR_FLAGS_AVP_CODE = convert_to_4_bytes(1443)
EQUIPMENT_STATUS_AVP_CODE = convert_to_4_bytes(1445)
RAND_AVP_CODE = convert_to_4_bytes(1447)
XRES_AVP_CODE = convert_to_4_bytes(1448)
AUTN_AVP_CODE = convert_to_4_bytes(1449)
KASME_AVP_CODE = convert_to_4_bytes(1450)
PDN_TYPE_AVP_CODE = convert_to_4_bytes(1456)
HOMOGENEOUS_SUPPORT_OF_IMS_VOICE_OVER_PS_SESSION_AVP_CODE = convert_to_4_bytes(1493)
ERROR_DIAGNOSTIC_AVP_CODE = convert_to_4_bytes(1614)
UE_SRVCC_CAPABILITY_AVP_CODE = convert_to_4_bytes(1615)
PUR_FLAGS_AVP_CODE = convert_to_4_bytes(1635)
CLR_FLAGS_AVP_CODE = convert_to_4_bytes(1638)
AIR_FLAGS_AVP_CODE = convert_to_4_bytes(1679)
UE_USAGE_TYPE_AVP_CODE = convert_to_4_bytes(1680)
SUPPORTED_SERVICES_AVP_CODE = convert_to_4_bytes(3143)
SUPPORTED_MONITORING_EVENTS_AVP_CODE = convert_to_4_bytes(3144)

#: List of Cancellation-Type AVP values.
#: For more information, please refer to Section 7.3.24 of 
#: ETSI TS 129 272 V15.4.0 (2018-07).
CANCELLATION_TYPE_MME_UPDATE_PROCEDURE = convert_to_4_bytes(0)
CANCELLATION_TYPE_SGSN_UPDATE_PROCEDURE = convert_to_4_bytes(1)
CANCELLATION_TYPE_SUBSCRIPTION_WITHDRAWAL = convert_to_4_bytes(2)
CANCELLATION_TYPE_UPDATE_PROCEDURE_IWF = convert_to_4_bytes(3)
CANCELLATION_TYPE_INITIAL_ATTACH_PROCEDURE = convert_to_4_bytes(4)

#: List of Subscriber-Status AVP values.
#: For more information, please refer to Section 7.3.29 of
#: ETSI TS 129 272 V15.4.0 (2018-07).
SUBSCRIBER_STATUS_SERVICE_GRANTED = convert_to_4_bytes(0)
SUBSCRIBER_STATUS_OPERATOR_DETERMINED_BARRING = convert_to_4_bytes(1)

#: List of All-APN-Configurations-Included-Indicator AVP values.
#: For more information, please refer to Section 7.3.33 of
#: ETSI TS 129 272 V15.4.0 (2018-07).
ALL_APN_CONFIGURATIONS_INCLUDED = convert_to_4_bytes(0)
MODIFIED_ADDED_APN_CONFIGURATIONS_INCLUDED = convert_to_4_bytes(1)

#: List of VPLMN-Dynamic-Address-Allowed.
#: For more information, please refer to Section 7.3.38 of 
#: ETSI TS 129 272 V12.6.0 (2014-10).
VPLMN_DYNAMIC_ADDRESS_ALLOWED_NOT_ALLOWED = convert_to_4_bytes(0)
VPLMN_DYNAMIC_ADDRESS_ALLOWED_ALLOWED = convert_to_4_bytes(1)

#: List of PDN-GW-Allocation-Type values.
#: For more information, please refer to Section 7.3.44 of
#  ETSI TS 129 272 V15.10.0 (2020-01).
PDN_GW_ALLOCATION_TYPE_STATIC = convert_to_4_bytes(0)
PDN_GW_ALLOCATION_TYPE_DYNAMIC = convert_to_4_bytes(1)

#: List of Equipment-Status AVP values.
#: For more information, please refer to Section 7.3.51 of 
#: ETSI TS 129 272 V15.4.0 (2018-07).
EQUIPMENT_STATUS_WHITELISTED = convert_to_4_bytes(0)
EQUIPMENT_STATUS_BLACKLISTED = convert_to_4_bytes(1)
EQUIPMENT_STATUS_GREYLISTED = convert_to_4_bytes(2)

#: List of PDN-Type values.
#: For more information, please refer to Section 7.3.62 of
#  ETSI TS 129 272 V15.10.0 (2020-01).
PDN_TYPE_IPV4 = convert_to_4_bytes(0)
PDN_TYPE_IPV6 = convert_to_4_bytes(1)
PDN_TYPE_IPV4V6 = convert_to_4_bytes(2)
PDN_TYPE_IPV4_OR_IPV6 = convert_to_4_bytes(3)

#: List of Alert-Reason AVP values.
#: For more information, please refer to Section 7.3.83 of
#: ETSI TS 129 272 V15.4.0 (2018-07).
ALERT_REASON_UE_PRESET = convert_to_4_bytes(0)
ALERT_REASON_UE_MEMORY_AVAILABLE = convert_to_4_bytes(1)

#: List of Homogeneous-Support-of-IMS-Voice-Over-PS-Sessions AVP values.
#: For more information, please refer to Section 7.3.107 of
#: ETSI TS 129 272 V15.4.0 (2018-07).
IMS_VOICE_OVER_PS_NOT_SUPPORTED = convert_to_4_bytes(0)
IMS_VOICE_OVER_PS_SUPPORTED = convert_to_4_bytes(1)

#: List of Error-Diagnostic AVP values.
#: For more information, please refer to Section 7.3.128 of
#: ETSI TS 129 272 V15.4.0 (2018-07).
ERROR_DIAGNOSTIC_GPRS_DATA_SUBSCRIBED = convert_to_4_bytes(0)
ERROR_DIAGNOSTIC_NO_GPRS_DATA_SUBSCRIBED = convert_to_4_bytes(1)
ERROR_DIAGNOSTIC_ODB_ALL_APN = convert_to_4_bytes(2)
ERROR_DIAGNOSTIC_ODB_HPLMN_APN = convert_to_4_bytes(3)
ERROR_DIAGNOSTIC_ODB_VPLMN_APN = convert_to_4_bytes(4)

#: List of UE-SRVCC-Capability AVP values.
#: For more information, please refer to Section 7.3.130 of 
#: ETSI TS 129 272 V15.10.0 (2020-01).
UE_SRVCC_NOT_SUPPORTED = convert_to_4_bytes(0)
UE_SRVCC_SUPPORTED = convert_to_4_bytes(1)



# -*- coding: utf-8 -*-
"""
    bromelia.constants.app_ids
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains list of Application Ids that are defined in several
    Diameter standard documents (both ETSI and IETF specs).
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""


from .._internal_utils import convert_to_4_bytes

DIAMETER_APPLICATION_DEFAULT = convert_to_4_bytes(0)
DIAMETER_APPLICATION_Cx = convert_to_4_bytes(16777216)
DIAMETER_APPLICATION_Sh = convert_to_4_bytes(16777217)
DIAMETER_APPLICATION_Zh = convert_to_4_bytes(16777221)
DIAMETER_APPLICATION_Rx = convert_to_4_bytes(16777236)
DIAMETER_APPLICATION_Gx = convert_to_4_bytes(16777238)
DIAMETER_APPLICATION_S6a = convert_to_4_bytes(16777251)
DIAMETER_APPLICATION_S6a_S6d = convert_to_4_bytes(16777251)
DIAMETER_APPLICATION_S13 = convert_to_4_bytes(16777252)
DIAMETER_APPLICATION_S13_S13 = convert_to_4_bytes(16777252)
DIAMETER_APPLICATION_SWm = convert_to_4_bytes(16777264)
DIAMETER_APPLICATION_SWx = convert_to_4_bytes(16777265)
DIAMETER_APPLICATION_S6b = convert_to_4_bytes(16777272)
DIAMETER_APPLICATION_SLh = convert_to_4_bytes(16777291)
DIAMETER_APPLICATION_UNKNOWN = convert_to_4_bytes(4123456789)
DIAMETER_APPLICATION_RELAY = convert_to_4_bytes(4294967295)



# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf.rfc8506
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    This module contains constants defined in IETF RFC 8506.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
CC_REQUEST_NUMBER_AVP_CODE = convert_to_4_bytes(415)
CC_REQUEST_TYPE_AVP_CODE = convert_to_4_bytes(416)
RATING_GROUP_AVP_CODE = convert_to_4_bytes(432)
SERVICE_IDENTIFIER_AVP_CODE = convert_to_4_bytes(439)
USER_EQUIPMENT_INFO_AVP_CODE = convert_to_4_bytes(458)
USER_EQUIPMENT_INFO_TYPE_AVP_CODE = convert_to_4_bytes(459)
USER_EQUIPMENT_INFO_VALUE_AVP_CODE = convert_to_4_bytes(460)

#: List of CC-Request-Type AVP values.
#: For more information, please refer to Section 8.3 of IETF RFC 8506.
CC_REQUEST_TYPE_INITIAL_REQUEST = convert_to_4_bytes(1)
CC_REQUEST_TYPE_UPDATE_REQUEST = convert_to_4_bytes(2)
CC_REQUEST_TYPE_TERMINATION_REQUEST = convert_to_4_bytes(3)
CC_REQUEST_TYPE_EVENT_REQUEST = convert_to_4_bytes(4)

#: List of User-Equipment-Info-Type AVP values.
#: For more information, please refer to Section 8.50 of IETF RFC 8506.
USER_EQUIPMENT_INFO_TYPE_IMEISV = convert_to_4_bytes(0)
USER_EQUIPMENT_INFO_TYPE_MAC = convert_to_4_bytes(1)
USER_EQUIPMENT_INFO_TYPE_EUI64 = convert_to_4_bytes(2)
USER_EQUIPMENT_INFO_TYPE_MODIFIED_EUI64 = convert_to_4_bytes(3)



# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_swm.avps
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library that are used 
    to create Diameter messages for 3GPP SWm Application Id.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...avps.etsi_3gpp.ts_129_061 import X3gppChargingCharacteristicsAVP

from ...avps.etsi_3gpp.ts_129_212 import RatTypeAVP
from ...avps.etsi_3gpp.ts_129_212 import UeLocalIpAddressAVP

from ...avps.etsi_3gpp.ts_129_229 import SupportedFeaturesAVP
from ...avps.etsi_3gpp.ts_129_229 import VisitedNetworkIdentifierAVP

from ...avps.etsi_3gpp.ts_129_272 import ServiceSelectionAVP
from ...avps.etsi_3gpp.ts_129_272 import ApnConfigurationAVP

from ...avps.etsi_3gpp.ts_129_273 import MobileNodeIdentifierAVP
from ...avps.etsi_3gpp.ts_129_273 import Mip6FeatureVectorAVP

from ...avps.ietf.rfc4006 import SubscriptionIdAVP

from ...avps.ietf.rfc4072 import EapMasterSessionKeyAVP
from ...avps.ietf.rfc4072 import EapPayloadAVP

from ...avps.ietf.rfc5447 import Mip6AgentInfoAVP

from ...avps.ietf.rfc6733 import AuthApplicationIdAVP
from ...avps.ietf.rfc6733 import AuthRequestTypeAVP
from ...avps.ietf.rfc6733 import AuthSessionStateAVP
from ...avps.ietf.rfc6733 import DestinationHostAVP
from ...avps.ietf.rfc6733 import DestinationRealmAVP
from ...avps.ietf.rfc6733 import OriginHostAVP
from ...avps.ietf.rfc6733 import OriginRealmAVP
from ...avps.ietf.rfc6733 import RedirectHostAVP
from ...avps.ietf.rfc6733 import ResultCodeAVP
from ...avps.ietf.rfc6733 import SessionIdAVP
from ...avps.ietf.rfc6733 import SessionTimeoutAVP
from ...avps.ietf.rfc6733 import UserNameAVP


# -*- coding: utf-8 -*-
"""
    bromelia.lib.ietf_rfc6733.avps
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol AVP library that are used 
    to create Diameter base messages.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...avps.ietf.rfc6733 import *


# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_129_214
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 129 214.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants.etsi_3gpp.ts_129_214 import *
from ...types import *


class MaxRequestedBandwidthDlAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Max-Requested-Bandwidth-DL AVP in Section 5.3.14 
    of ETSI TS 129 214 V15.4.0 (2018-07).

    The Max-Requested-Bandwidth-DL AVP (AVP Code 515) is of type Unsigned32.
    """
    code = MAX_REQUESTED_BANDWIDTH_DL_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             MaxRequestedBandwidthDlAVP.code,
                             MaxRequestedBandwidthDlAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class MaxRequestedBandwidthUlAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Max-Requested-Bandwidth-UL AVP in Section 5.3.15 
    of ETSI TS 129 214 V15.4.0 (2018-07).

    The Max-Requested-Bandwidth-UL AVP (AVP Code 516) is of type Unsigned32.
    """
    code = MAX_REQUESTED_BANDWIDTH_UL_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             MaxRequestedBandwidthUlAVP.code,
                             MaxRequestedBandwidthUlAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AbortCauseAVP(DiameterAVP, EnumeratedType):
    """Implementation of Abort-Cause AVP in Section 5.3.1 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Abort-Cause AVP (AVP Code 500) is of type Enumerated.
    """
    code = ABORT_CAUSE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                ABORT_CAUSE_BEARER_RELEASED,
                ABORT_CAUSE_INSUFFICIENT_SERVER_RESOURCES,
                ABORT_CAUSE_INSUFFICIENT_BEARER_RESOURCES,
                ABORT_CAUSE_PS_TO_CS_HANDOVER,
                ABORT_CAUSE_SPONSORED_DATA_CONNECTIVITY_DISALLOWED,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AbortCauseAVP.code,
                             AbortCauseAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AfApplicationIdentifierAVP(DiameterAVP, OctetStringType):
    """Implementation of AF-Application-identifier AVP in Section 5.3.5  
    of ETSI TS 129 214 V15.4.0 (2018-07).

    The AF-Application-identifier AVP (AVP Code 504) is of type OctetString.
    """
    code = AF_APPLICATION_IDENTIFIER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AfApplicationIdentifierAVP.code,
                             AfApplicationIdentifierAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AfChargingIdentifierAVP(DiameterAVP, OctetStringType):
    """Implementation of AF-Charging-identifier AVP in Section 5.3.6  
    of ETSI TS 129 214 V15.4.0 (2018-07).

    The AF-Charging-identifier AVP (AVP Code 505) is of type OctetString.
    """
    code = AF_CHARGING_IDENTIFIER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AfChargingIdentifierAVP.code,
                             AfChargingIdentifierAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class FlowDescriptionAVP(DiameterAVP, OctetStringType):
    """Implementation of Flow-Description AVP in Section 5.38 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Flow-Description AVP (AVP Code 507) is of type IPFilterRule.
    """
    code = FLOW_DESCRIPTION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             FlowDescriptionAVP.code,
                             FlowDescriptionAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class FlowNumberAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Flow-Number AVP in Section 5.3.9 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Flow-Number AVP (AVP Code 509) is of type Unsigned32.
    """
    code = FLOW_NUMBER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             FlowNumberAVP.code,
                             FlowNumberAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)
 

class FlowStatusAVP(DiameterAVP, EnumeratedType):
    """Implementation of Flow-Status AVP in Section 5.3.11 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Flow-Status AVP (AVP Code 511) is of type Enumerated.
    """
    code = FLOW_STATUS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                FLOW_STATUS_ENABLED_UPLINK,
                FLOW_STATUS_ENABLED_DOWNLINK,
                FLOW_STATUS_ENABLED,
                FLOW_STATUS_DISABLED,
                FLOW_STATUS_REMOVED,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             FlowStatusAVP.code,
                             FlowStatusAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class FlowUsageAVP(DiameterAVP, EnumeratedType):
    """Implementation of Flow-Usage AVP in Section 5.3.12 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Flow-Usage AVP (AVP Code 512) is of type Enumerated.
    """
    code = FLOW_USAGE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                FLOW_USAGE_NO_INFORMATION,
                FLOW_USAGE_RTCP,
                FLOW_USAGE_AF_SIGNALLING,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             FlowUsageAVP.code,
                             FlowUsageAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class SpecificActionAVP(DiameterAVP, EnumeratedType):
    """Implementation of Specific-Action AVP in Section 5.3.13 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Specific-Action AVP (AVP Code 513) is of type Enumerated.
    """
    code = SPECIFIC_ACTION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    values = [
                SPECIFIC_ACTION_CHARGING_CORRELATION_EXCHANGE,
                SPECIFIC_ACTION_INDICATION_OF_LOSS_OF_BEARER,
                SPECIFIC_ACTION_INDICATION_OF_RECOVERY_OF_BEARER,
                SPECIFIC_ACTION_INDICATION_OF_RELEASE_OF_BEARER,
                SPECIFIC_ACTION_IP_CAN_CHANGE,
                SPECIFIC_ACTION_INDICATION_OF_OUT_OF_CREDIT,
                SPECIFIC_ACTION_INDICATION_OF_SUCCESSFUL_RESOURCES_ALLOCATION,
                SPECIFIC_ACTION_INDICATION_OF_FAILED_RESOURCES_ALLOCATION,
                SPECIFIC_ACTION_INDICATION_OF_LIMITED_PCC_DEPLOYMENT,
                SPECIFIC_ACTION_USAGE_REPORT,
                SPECIFIC_ACTION_ACCESS_NETWORK_INFO_REPORT,
                SPECIFIC_ACTION_INDICATION_OF_RECOVERY_FROM_LIMITED_PCC_DEPLOYMENT,
                SPECIFIC_ACTION_INDICATION_OF_ACCESS_NETWORK_INFO_REPORTING_FAILURE,
                SPECIFIC_ACTION_INDICATION_OF_TRANSFER_POLICY_EXPIRED,
                SPECIFIC_ACTION_PLMN_CHANGE,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             SpecificActionAVP.code,
                             SpecificActionAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class MediaComponentDescriptionAVP(DiameterAVP, GroupedType):
    """Implementation of Media-Component-Description AVP in Section 5.3.16 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Media-Component-Description AVP (AVP Code 517) is of type Grouped.
    """
    code = MEDIA_COMPONENT_DESCRIPTION_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):        
        DiameterAVP.__init__(self, 
                             MediaComponentDescriptionAVP.code,
                             MediaComponentDescriptionAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class MediaComponentNumberAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Media-Component-Number AVP in Section 5.3.17 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Media-Component-Number AVP (AVP Code 518) is of type Unsigned32.
    """
    code = MEDIA_COMPONENT_NUMBER_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             MediaComponentNumberAVP.code,
                             MediaComponentNumberAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        Unsigned32Type.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class MediaSubComponentAVP(DiameterAVP, GroupedType):
    """Implementation of Media-Sub-Component AVP in Section 5.3.18 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Media-Sub-Component AVP (AVP Code 519) is of type Grouped.
    """
    code = MEDIA_SUB_COMPONENT_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):        
        DiameterAVP.__init__(self, 
                             MediaSubComponentAVP.code,
                             MediaSubComponentAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        GroupedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class MediaTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Media-Type AVP in Section 5.3.19 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Media-Type AVP (AVP Code 520) is of type Enumerated.
    """
    code = MEDIA_TYPE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP
 
    values = [
                MEDIA_TYPE_AUDIO,
                MEDIA_TYPE_VIDEO,
                MEDIA_TYPE_DATA,
                MEDIA_TYPE_APPLICATION,
                MEDIA_TYPE_CONTROL,
                MEDIA_TYPE_TEXT,
                MEDIA_TYPE_MESSAGE,
                MEDIA_TYPE_OTHER,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             MediaTypeAVP.code,
                             MediaTypeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class ServiceInfoStatusAVP(DiameterAVP, EnumeratedType):
    """Implementation of Service-Info-Status AVP in Section 5.3.25 of 
    ETSI TS 129 214 V15.4.0 (2018-07).

    The Service-Info-Status AVP (AVP Code 527) is of type Enumerated.
    """
    code = SERVICE_INFO_STATUS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP
 
    values = [
                SERVICE_INFO_STATUS_FINAL_SERVICE_INFORMATION,
                SERVICE_INFO_STATUS_PRELIMINARY_SERVICE_INFORMATION,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ServiceInfoStatusAVP.code,
                             ServiceInfoStatusAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        EnumeratedType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AccessNetworkChargingAddressAVP(DiameterAVP, AddressType):
    """Implementation of Access-Network-Charging-Address AVP in Section 5.3.2 of 
    ETSI 3GPP TS 29.214 V8.14.0 (2012-12).

    The Access-Network-Charging-Address AVP (AVP Code 501) is of type Address.
    """
    code = ACCESS_NETWORK_CHARGING_ADDRESS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AccessNetworkChargingAddressAVP.code,
                             AccessNetworkChargingAddressAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        AddressType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


class AccessNetworkChargingIdentifierValueAVP(DiameterAVP, OctetStringType):
    """Implementation of Access-Network-Charging-Identifier-Value AVP in 
    Section 5.3.4 of ETSI 3GPP TS 29.214 V8.14.0 (2012-12).

    The Access-Network-Charging-Identifier-Value AVP (AVP Code 503) is of 
    type OctetString.
    """
    code = ACCESS_NETWORK_CHARGING_IDENTIFIER_VALUE_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             AccessNetworkChargingIdentifierValueAVP.code,
                             AccessNetworkChargingIdentifierValueAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterAVP.set_vendor_id_bit(self, True)
        OctetStringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)



# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf.rfc7155
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    
    
    This module contains constants defined in IETF RFC 7155.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
FRAMED_IP_ADDRESS_AVP_CODE = convert_to_4_bytes(8)
CALLED_STATION_ID_AVP_CODE = convert_to_4_bytes(30)
CALLING_STATION_ID_AVP_CODE = convert_to_4_bytes(31)
FRAMED_IPV6_PREFIX_AVP_CODE = convert_to_4_bytes(97)



# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp_rx
    ~~~~~~~~~~~~~~~~~~~~~

    A Python package with APIs for handling 3GPP Rx Application Id in the
    bromelia Python Library context.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""


# -*- coding: utf-8 -*-
"""
    bromelia.constants.etsi_3gpp.ts_129_212
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in ETSI TS 129 212.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_3_bytes
from ..._internal_utils import convert_to_4_bytes


#: Diameter Messages
CC_MESSAGE = convert_to_3_bytes(272)

#: Diameter AVPs
BEARER_USAGE_AVP_CODE = convert_to_4_bytes(1000)
CHARGING_RULE_INSTALL_AVP_CODE = convert_to_4_bytes(1001)
CHARGING_RULE_DEFINITION_AVP_CODE = convert_to_4_bytes(1003)
CHARGING_RULE_NAME_AVP_CODE = convert_to_4_bytes(1005)
EVENT_TRIGGER_AVP_CODE = convert_to_4_bytes(1006)
METERING_METHOD_AVP_CODE = convert_to_4_bytes(1007)
OFFLINE_AVP_CODE = convert_to_4_bytes(1008)
ONLINE_AVP_CODE = convert_to_4_bytes(1009)
PRECEDENCE_AVP_CODE = convert_to_4_bytes(1010)
REPORTING_LEVEL_AVP_CODE = convert_to_4_bytes(1011)
QOS_INFORMATION_AVP_CODE = convert_to_4_bytes(1016)
CHARGING_RULE_REPORT_AVP_CODE = convert_to_4_bytes(1018)
PCC_RULE_STATUS_AVP_CODE = convert_to_4_bytes(1019)
ACCESS_NETWORK_CHARGING_IDENTIFIER_GX_AVP_CODE = convert_to_4_bytes(1022)
NETWORK_REQUEST_SUPPORT_AVP_CODE = convert_to_4_bytes(1024)
GUARANTEED_BITRATE_DL_AVP_CODE = convert_to_4_bytes(1025)
GUARANTEED_BITRATE_UL_AVP_CODE = convert_to_4_bytes(1026)
IP_CAN_TYPE_AVP_CODE = convert_to_4_bytes(1027)
QOS_CLASS_IDENTIFIER_AVP_CODE = convert_to_4_bytes(1028)
RULE_FAILURE_CODE_AVP_CODE = convert_to_4_bytes(1031)
RAT_TYPE_AVP_CODE = convert_to_4_bytes(1032)
APN_AGGREGATE_MAX_BITRATE_DL_AVP_CODE = convert_to_4_bytes(1040)
APN_AGGREGATE_MAX_BITRATE_UL_AVP_CODE = convert_to_4_bytes(1041)
PRIORITY_LEVEL_AVP_CODE = convert_to_4_bytes(1046)
PRE_EMPTION_CAPABILITY_AVP_CODE = convert_to_4_bytes(1047)
PRE_EMPTION_VULNERABILITY_AVP_CODE = convert_to_4_bytes(1048)
DEFAULT_EPS_BEARER_QOS_AVP_CODE = convert_to_4_bytes(1049)
AN_GW_ADDRESS_AVP_CODE = convert_to_4_bytes(1050)
FLOW_INFORMATION_AVP_CODE = convert_to_4_bytes(1058)
CHARGING_CORRELATION_INDICATOR_AVP_CODE = convert_to_4_bytes(1073)
UE_LOCAL_IP_ADDRESS_AVP_CODE = convert_to_4_bytes(2805)

#: List of Bearer-Usage AVP values.
#: For more information, please refer to Section 5.3.1 of 
# ETSI TS 129 212 V15.3.0 (2018-07).
BEARER_USAGE_GENERAL = convert_to_4_bytes(0)
BEARER_USAGE_SIGNALLING = convert_to_4_bytes(1)

#: List of Event-Trigger AVP values.
#: For more information, please refer to Section 5.3.7 of 
#: ETSI TS 129 212 V15.3.0 (2018-07).
EVENT_TRIGGER_SGSN_CHANGE = convert_to_4_bytes(0)
EVENT_TRIGGER_QOS_CHANGE = convert_to_4_bytes(1)
EVENT_TRIGGER_RAT_CHANGE = convert_to_4_bytes(2)
EVENT_TRIGGER_TFT_CHANGE = convert_to_4_bytes(3)
EVENT_TRIGGER_PLMN_CHANGE = convert_to_4_bytes(4)
EVENT_TRIGGER_LOSS_OF_BEARER = convert_to_4_bytes(5)
EVENT_TRIGGER_RECOVERY_OF_BEARER = convert_to_4_bytes(6)
EVENT_TRIGGER_IP_CAN_CHANGE = convert_to_4_bytes(7)
EVENT_TRIGGER_QOS_CHANGE_EXCEEDING_AUTHORIZATION = convert_to_4_bytes(11)
EVENT_TRIGGER_RAI_CHANGE = convert_to_4_bytes(12)
EVENT_TRIGGER_USER_LOCATION_CHANGE = convert_to_4_bytes(13)
EVENT_TRIGGER_NO_EVENT_TRIGGERS = convert_to_4_bytes(14)
EVENT_TRIGGER_OUT_OF_CREDIT = convert_to_4_bytes(15)
EVENT_TRIGGER_REALLOCATION_OF_CREDIT = convert_to_4_bytes(16)
EVENT_TRIGGER_REVALIDATION_TIMEOUT = convert_to_4_bytes(17)
EVENT_TRIGGER_UE_IP_ADDRESS_ALLOCATE = convert_to_4_bytes(18)
EVENT_TRIGGER_UE_IP_ADDRESS_RELEASE = convert_to_4_bytes(19)
EVENT_TRIGGER_DEFAULT_EPS_BEARER_QOS_CHANGE = convert_to_4_bytes(20)
EVENT_TRIGGER_AN_GW_CHANGE = convert_to_4_bytes(21)
EVENT_TRIGGER_SUCCESSFUL_RESOURCE_ALLOCATION = convert_to_4_bytes(22)
EVENT_TRIGGER_RESOURCE_MODIFICATION_REQUEST = convert_to_4_bytes(23)
EVENT_TRIGGER_PGW_TRACE_CONTROL = convert_to_4_bytes(24)
EVENT_TRIGGER_UE_TIME_ZONE_CHANGE = convert_to_4_bytes(25)
EVENT_TRIGGER_TAI_CHANGE = convert_to_4_bytes(26)
EVENT_TRIGGER_ECGI_CHANGE = convert_to_4_bytes(27)
EVENT_TRIGGER_CHARGING_CORRELATION_EXCHANGE = convert_to_4_bytes(28)
EVENT_TRIGGER_APN_AMBR_MODIFICATION_FAILURE = convert_to_4_bytes(29)
EVENT_TRIGGER_USER_CSG_INFORMATION_CHANGE = convert_to_4_bytes(30)
EVENT_TRIGGER_USAGE_REPORT = convert_to_4_bytes(33)
EVENT_TRIGGER_DEFAULT_EPS_BEARER_QOS_MODIFICATION_FAILURE = convert_to_4_bytes(34)
EVENT_TRIGGER_USER_CSG_HYBRID_SUBSCRIBED_INFORMATION_CHANGE = convert_to_4_bytes(35)
EVENT_TRIGGER_USER_CSG_HYBRID_UNSUBSCRIBED_INFORMATION_CHANGE = convert_to_4_bytes(36)
EVENT_TRIGGER_ROUTING_RULE_CHANGE = convert_to_4_bytes(37)
EVENT_TRIGGER_APPLICATION_START = convert_to_4_bytes(39)
EVENT_TRIGGER_APPLICATION_STOP = convert_to_4_bytes(40)
EVENT_TRIGGER_CS_TO_PS_HANDOVER = convert_to_4_bytes(42)
EVENT_TRIGGER_UE_LOCAL_IP_ADDRESS_CHANGE = convert_to_4_bytes(43)
EVENT_TRIGGER_HENB_LOCAL_IP_ADDRESS_CHANGE = convert_to_4_bytes(44)
EVENT_TRIGGER_ACCESS_NETWORK_INFO_REPORT = convert_to_4_bytes(45)
EVENT_TRIGGER_CREDIT_MANAGEMENT_SESSION_FAILURE = convert_to_4_bytes(46)
EVENT_TRIGGER_DEFAULT_QOS_CHANGE = convert_to_4_bytes(47)
EVENT_TRIGGER_CHANGE_OF_UE_PRESENCE_IN_PRESENCE_REPORTING_AREA_REPORT = convert_to_4_bytes(48)
EVENT_TRIGGER_ADDITION_OF_ACCESS = convert_to_4_bytes(49)
EVENT_TRIGGER_REMOVAL_OF_ACCESS = convert_to_4_bytes(50)
EVENT_TRIGGER_UNAVAILABILITY_OF_ACCESS = convert_to_4_bytes(51)
EVENT_TRIGGER_AVAILABILITY_OF_ACCESS = convert_to_4_bytes(52)
EVENT_TRIGGER_RESOURCE_RELEASE = convert_to_4_bytes(53)
EVENT_TRIGGER_ENODEB_CHANGE = convert_to_4_bytes(54)
EVENT_TRIGGER_3GPP_PS_DATA_OFF_CHANGE = convert_to_4_bytes(55)

#: List of Metering-Method AVP values.
#: For more information, please refer to Section 5.3.8 of
#: ETSI TS 129 212 V15.3.0 (2018-07).
METERING_METHOD_DURATION = convert_to_4_bytes(0)
METERING_METHOD_VOLUME = convert_to_4_bytes(1)
METERING_METHOD_DURATION_VOLUME = convert_to_4_bytes(2)
METERING_METHOD_EVENT = convert_to_4_bytes(3)

#: List of Offline AVP values.
#: For more information, please refer to Section 5.3.9 of
#: ETSI TS 129 212 V15.3.0 (2018-07).
OFFLINE_DISABLE_OFFLINE = convert_to_4_bytes(0)
OFFLINE_ENABLE_OFFLINE = convert_to_4_bytes(1)

#: List of Online AVP values.
#: For more information, please refer to Section 5.3.10 of
#: ETSI TS 129 212 V15.3.0 (2018-07).
ONLINE_DISABLE_ONLINE = convert_to_4_bytes(0)
ONLINE_ENABLE_ONLINE = convert_to_4_bytes(1)

#: List of Reporting-Level AVP values.
#: For more information, please refer to Section 5.3.12 of 
#: ETSI TS 129 212 V15.3.0 (2018-07).
REPORTING_LEVEL_SERVICE_IDENTIFIER_LEVEL = convert_to_4_bytes(0)
REPORTING_LEVEL_RATING_GROUP_LEVEL = convert_to_4_bytes(1)
REPORTING_LEVEL_SPONSORED_CONNECTIVITY_LEVEL = convert_to_4_bytes(2)

#: List of QOS-CLASS-IDENTIFIER.
#: For more information, please refer to Section 5.3.17 of 
#: of ETSI TS 129 212 V15.3.0 (2018-07).
QCI_1 = convert_to_4_bytes(1)
QCI_2 = convert_to_4_bytes(2)
QCI_3 = convert_to_4_bytes(3)
QCI_4 = convert_to_4_bytes(4)
QCI_5 = convert_to_4_bytes(5)
QCI_6 = convert_to_4_bytes(6)
QCI_7 = convert_to_4_bytes(7)
QCI_8 = convert_to_4_bytes(8)
QCI_9 = convert_to_4_bytes(9)
QCI_65 = convert_to_4_bytes(65)
QCI_66 = convert_to_4_bytes(66)
QCI_69 = convert_to_4_bytes(69)
QCI_70 = convert_to_4_bytes(70)
QCI_75 = convert_to_4_bytes(75)
QCI_79 = convert_to_4_bytes(79)
QCI_80 = convert_to_4_bytes(80)
QCI_82 = convert_to_4_bytes(82)
QCI_83 = convert_to_4_bytes(83)
QCI_84 = convert_to_4_bytes(84)
QCI_85 = convert_to_4_bytes(85)

#: List of PCC-Rule-Status AVP values.
#: For more information, please refer to Section 5.3.19 of
#: ETSI TS 129 212 V15.3.0 (2018-07).
PCC_RULE_STATUS_ACTIVE = convert_to_4_bytes(0)
PCC_RULE_STATUS_INACTIVE = convert_to_4_bytes(1)
PCC_RULE_STATUS_TEMPORARILY_INACTIVE = convert_to_4_bytes(2)

#: List of Network-Request-Support AVP values.
#: For more information, please refer to Section 5.3.24 of
#: ETSI TS 129 212 V15.3.0 (2018-07).
NETWORK_REQUEST_NOT_SUPPORTED = convert_to_4_bytes(0)
NETWORK_REQUEST_SUPPORTED = convert_to_4_bytes(1)

#: List of IP-CAN-Type AVP values.
#: For more information, please refer to Section 5.3.27 of 
#: ETSI TS 129 212 V15.3.0 (2018-07).
IP_CAN_TYPE_3GPP_GPRS = convert_to_4_bytes(0)
IP_CAN_TYPE_DOCSIS = convert_to_4_bytes(1)
IP_CAN_TYPE_XDSL = convert_to_4_bytes(2)
IP_CAN_TYPE_WIMAX = convert_to_4_bytes(3)
IP_CAN_TYPE_3GPP2 = convert_to_4_bytes(4)
IP_CAN_TYPE_3GPP_EPS = convert_to_4_bytes(5)
IP_CAN_TYPE_NON_3GPP_EPS = convert_to_4_bytes(6)
IP_CAN_TYPE_FBA = convert_to_4_bytes(7)
IP_CAN_TYPE_3GPP_5GS = convert_to_4_bytes(8)
IP_CAN_TYPE_NON_3GPP_5GS = convert_to_4_bytes(9)

#: List of RAT-TYPES.
#: For more information, please refer to Section 5.3.31 of 3GPP TS 129 212.
RAT_TYPE_WLAN = convert_to_4_bytes(0)
RAT_TYPE_VIRTUAL = convert_to_4_bytes(1)
RAT_TYPE_UTRAN = convert_to_4_bytes(1000)
RAT_TYPE_GERAN = convert_to_4_bytes(1001)
RAT_TYPE_GAN = convert_to_4_bytes(1002)
RAT_TYPE_HSPA_EVOLUTION = convert_to_4_bytes(1003)
RAT_TYPE_EUTRAN = convert_to_4_bytes(1004)
RAT_TYPE_CDMA2000_1X = convert_to_4_bytes(2000)
RAT_TYPE_HRPD = convert_to_4_bytes(2001)
RAT_TYPE_UMB = convert_to_4_bytes(2002)
RAT_TYPE_EHRPD = convert_to_4_bytes(2003)

#: List of Rule-Failure-Code AVP values.
#: For more information, please refer to Section 5.3.38 of
#: ETSI TS 129 212 V15.3.0 (2018-07).
RULE_FAILURE_CODE_UNKNOWN_RULE_NAME = convert_to_4_bytes(1)
RULE_FAILURE_CODE_RATING_GROUP_ERROR = convert_to_4_bytes(2)
RULE_FAILURE_CODE_SERVICE_IDENTIFIER_ERROR = convert_to_4_bytes(3)
RULE_FAILURE_CODE_GW_PCEF_MALFUNCTION = convert_to_4_bytes(4)
RULE_FAILURE_CODE_RESOURCES_LIMITATION = convert_to_4_bytes(5)
RULE_FAILURE_CODE_MAX_NR_BEARERS_REACHED = convert_to_4_bytes(6)
RULE_FAILURE_CODE_UNKNOWN_BEARER_ID = convert_to_4_bytes(7)
RULE_FAILURE_CODE_MISSING_BEARER_ID = convert_to_4_bytes(8)
RULE_FAILURE_CODE_MISSING_FLOW_INFORMATION = convert_to_4_bytes(9)
RULE_FAILURE_CODE_RESOURCE_ALLOCATION_FAILURE = convert_to_4_bytes(10)
RULE_FAILURE_CODE_UNSUCCESSFUL_QOS_VALIDATION = convert_to_4_bytes(11)
RULE_FAILURE_CODE_INCORRECT_FLOW_INFORMATION = convert_to_4_bytes(12)
RULE_FAILURE_CODE_PS_TO_CS_HANDOVER = convert_to_4_bytes(13)
RULE_FAILURE_CODE_TDF_APPLICATION_IDENTIFIER_ERROR = convert_to_4_bytes(14)
RULE_FAILURE_CODE_NO_BEARER_BOUND = convert_to_4_bytes(15)
RULE_FAILURE_CODE_FILTER_RESTRICTIONS = convert_to_4_bytes(16)
RULE_FAILURE_CODE_AN_GW_FAILED = convert_to_4_bytes(17)
RULE_FAILURE_CODE_MISSING_REDIRECT_SERVER_ADDRESS = convert_to_4_bytes(18)
RULE_FAILURE_CODE_CM_END_USER_SERVICE_DENIED = convert_to_4_bytes(19)
RULE_FAILURE_CODE_CM_CREDIT_CONTROL_NOT_APPLICABLE = convert_to_4_bytes(20)
RULE_FAILURE_CODE_CM_AUTHORIZATION_REJECTED = convert_to_4_bytes(21)
RULE_FAILURE_CODE_CM_USER_UNKNOWN = convert_to_4_bytes(22)
RULE_FAILURE_CODE_CM_RATING_FAILED = convert_to_4_bytes(23)
RULE_FAILURE_CODE_ROUTING_RULE_REJECTION = convert_to_4_bytes(24)
RULE_FAILURE_CODE_UNKNOWN_ROUTING_ACCESS_INFORMATION = convert_to_4_bytes(25)
RULE_FAILURE_CODE_NO_NBIFOM_SUPPORT = convert_to_4_bytes(26)

#: List of Pre-emption-Capability AVP values.
#: For more information, please refer to Section Section 5.3.46 of 
#: ETSI TS 129 212 V15.3.0 (2018-07).
PRE_EMPTION_CAPABILITY_ENABLED = convert_to_4_bytes(0)
PRE_EMPTION_CAPABILITY_DISABLED = convert_to_4_bytes(1)

#: List of Pre-emption-Vulnerability AVP values.
#: For more information, please refer to Section Section 5.3.47 of 
#: ETSI TS 129 212 V15.3.0 (2018-07).
PRE_EMPTION_VULNERABILITY_ENABLED = convert_to_4_bytes(0)
PRE_EMPTION_VULNERABILITY_DISABLED = convert_to_4_bytes(1)

#: List of Charging-Correlation-Indicator AVP values.
#: For more information, please refer to Section 5.3.67 of 
#: ETSI TS 129 212 V15.3.0 (2018-07).
CHARGING_CORRELATION_INDICATOR_CHARGING_IDENTIFIER_REQUIRED = convert_to_4_bytes(0)



# -*- coding: utf-8 -*-
"""
    bromelia.proxy
    ~~~~~~~~~~~~~~

    This module contains the BaseMessages class necessary to instantiate
    the DiameterAssociation class. It allows define the set of base messages
    of Diameter protocol which will be used in a specific connection. Besides
    that, it allows to customize the base messages content as per RFC 6733.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .constants import *
from .avps import AuthApplicationIdAVP
from .avps import VendorIdAVP
from .avps import VendorSpecificApplicationIdAVP
from .messages import CEA
from .messages import CER
from .messages import DWA
from .messages import DWR
from .messages import DPA
from .messages import DPR

class BaseMessages:
    def __init__(self, cer, cea, dwr, dwa, dpr, dpa):
        self.cer = cer
        self.cea = cea
        self.dwr = dwr
        self.dwa = dwa
        self.dpr = dpr
        self.dpa = dpa


class DiameterBaseProxy:
    def __init__(self, connection):
        self.connection = connection


    def get_default_messages(self):
        return BaseMessages(cer=DiameterBaseProxy.load_cer(self.connection),
                            cea=DiameterBaseProxy.load_cea(self.connection), 
                            dwr=DiameterBaseProxy.load_dwr(self.connection), 
                            dwa=DiameterBaseProxy.load_dwa(self.connection), 
                            dpr=DiameterBaseProxy.load_dpr(self.connection), 
                            dpa=DiameterBaseProxy.load_dpa(self.connection))


    def get_custom_messages(self, msgs):
        base = dict()
        for msg in msgs:
            if isinstance(msg, CER):
                base.update({"cer": msg})
            elif isinstance(msg, CEA):
                base.update({"cea": msg})
            elif isinstance(msg, DWR):
                base.update({"dwr": msg})
            elif isinstance(msg, DWA):
                base.update({"dwa": msg})
            elif isinstance(msg, DPR):
                base.update({"dpr": msg})
            elif isinstance(msg, DPA):
                base.update({"dpa": msg})

        for key in base.keys():
            if key == "cer":
                base.update({"cer": DiameterBaseProxy.load_cer(self.connection)})
            elif key == "cea":
                base.update({"cea": DiameterBaseProxy.load_cea(self.connection)})
            elif key == "dwr":
                base.update({"dwr": DiameterBaseProxy.load_dwr(self.connection)})
            elif key == "dwa":
                base.update({"dwa": DiameterBaseProxy.load_dwa(self.connection)})
            elif key == "dpr":
                base.update({"dpr": DiameterBaseProxy.load_dpr(self.connection)})
            elif key == "dpa":
                base.update({"dpa": DiameterBaseProxy.load_dpa(self.connection)})
        
        return BaseMessages(**base)


    @staticmethod
    def load_cer(connection):
        if isinstance(connection.application_ids, dict):
            application = connection.application_ids
            avps = {
                    "origin_host": connection.local_node.host_name,
                    "origin_realm": connection.local_node.realm,
                    "host_ip_address": connection.local_node.ip_address,
                    "vendor_id": VENDOR_ID_DEFAULT,
                    "auth_application_id": application["app_id"]
            }
            cer = CER(**avps)

        elif isinstance(connection.application_ids, list):
            avps = {
                    "origin_host": connection.local_node.host_name,
                    "origin_realm": connection.local_node.realm,
                    "host_ip_address": connection.local_node.ip_address,
                    "vendor_id": VENDOR_ID_DEFAULT,
            }
            cer = CER(**avps)

            applications = connection.application_ids
            for application in applications:
                vendor_id_avp = VendorIdAVP(application["vendor_id"])
                auth_app_id_avp = AuthApplicationIdAVP(application["app_id"])

                avp = VendorSpecificApplicationIdAVP([
                                                        vendor_id_avp, 
                                                        auth_app_id_avp
                ])

                cer.append(avp)

        return cer

    @staticmethod
    def load_cea(connection):
        if isinstance(connection.application_ids, dict):
            application = connection.application_ids
            avps = {
                    "origin_host": connection.local_node.host_name,
                    "origin_realm": connection.local_node.realm,
                    "host_ip_address": connection.local_node.ip_address,
                    "vendor_id": VENDOR_ID_DEFAULT,
                    "auth_application_id": application["app_id"],
            }
            cea = CEA(**avps)

        elif isinstance(connection.application_ids, list):
            avps = {
                    "origin_host": connection.local_node.host_name,
                    "origin_realm": connection.local_node.realm,
                    "host_ip_address": connection.local_node.ip_address,
                    "vendor_id": VENDOR_ID_DEFAULT,
            }
            cea = CEA(**avps)

            applications = connection.application_ids
            for application in applications:
                vendor_id_avp = VendorIdAVP(application["vendor_id"])
                auth_app_id_avp = AuthApplicationIdAVP(application["app_id"])

                avp = VendorSpecificApplicationIdAVP([
                                                        vendor_id_avp, 
                                                        auth_app_id_avp
                ])

                cea.append(avp)

        return cea


    @staticmethod
    def load_dwr(connection):
        avps = {
                "origin_host": connection.local_node.host_name,
                "origin_realm": connection.local_node.realm,
        }            
        return DWR(**avps)


    @staticmethod
    def load_dwa(connection):
        avps = {
                "origin_host": connection.local_node.host_name,
                "origin_realm": connection.local_node.realm,
        }
        return DWA(**avps)


    @staticmethod
    def load_dpr(connection):
        avps = {
                "origin_host": connection.local_node.host_name,
                "origin_realm": connection.local_node.realm,
        }            
        return DPR(**avps)


    @staticmethod
    def load_dpa(connection):
        avps = {
                "origin_host": connection.local_node.host_name,
                "origin_realm": connection.local_node.realm,
        }
        return DPA(**avps)



# -*- coding: utf-8 -*-
"""
    bromelia.constants.etsi_3gpp.ts_183_017
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in ETSI TS 183 017.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
RESERVATION_PRIORITY_AVP_CODE = convert_to_4_bytes(458)

#: List of Reservation-Priority AVP values.
#: For more information, please refer to Section 7.3.9 of 
#: ETSI TS 183 017 V2.3.1 (2008-09).
PRIORITY_DEFAULT = convert_to_4_bytes(0)
PRIORITY_ONE = convert_to_4_bytes(1)
PRIORITY_TWO = convert_to_4_bytes(2)
PRIORITY_THREE = convert_to_4_bytes(3)
PRIORITY_FOUR = convert_to_4_bytes(4)
PRIORITY_FIVE = convert_to_4_bytes(5)
PRIORITY_SIX = convert_to_4_bytes(6)
PRIORITY_SEVEN = convert_to_4_bytes(7)
